import xtrack as xt
import numpy as np

from scipy.constants import hbar
from scipy.constants import electron_volt
from scipy.constants import c as clight

env = xt.load_madx_lattice('b075_2024.09.25.madx')
line = env.ring
line.particle_ref = xt.Particles(energy0=2.7e9, mass0=xt.ELECTRON_MASS_EV)
line.configure_bend_model(num_multipole_kicks=20)

line['vrf'] = 1.8e6
line['frf'] = 499.6e6
line['lagrf'] = 180.

line.insert(
    env.new('cav', 'Cavity', voltage='vrf', frequency='frf', lag='lagrf', at=0))

tt = line.get_table()
tw4d_thick = line.twiss4d()
tw6d_thick = line.twiss()

line.build_tracker()
line.configure_radiation(model='mean')

tt_bend = tt.rows[tt.element_type == 'Bend']
tt_quad = tt.rows[tt.element_type == 'Quadrupole']
for nn in tt_bend.name:
    line.get(nn).model = 'mat-kick-mat'
    line.get(nn).integrator = 'yoshida4'
    line[nn].num_multipole_kicks = 20
# # for nn in tt_quad.name:
# #     line[nn].radiation_flag = 0

tw_rad_thick = line.twiss(eneloss_and_damping=True, strengths=True)

print('Done thick')

# Sliced thick
env['ring_sliced_thick'] = env.ring.copy(shallow=True)
line_sliced_thick = env['ring_sliced_thick']

line_sliced_thick.discard_tracker()
slicing_strategies = [
    xt.Strategy(slicing=None),  # Default
    xt.Strategy(slicing=xt.Teapot(2, mode='thick'), element_type=xt.Bend),
    xt.Strategy(slicing=xt.Teapot(2, mode='thick'), element_type=xt.Quadrupole),
]
line_sliced_thick.slice_thick_elements(slicing_strategies)

assert line['ars12_mqua_5940..0'].radiation_flag == 10

tw_rad_sliced_thick = line_sliced_thick.twiss(eneloss_and_damping=True, strengths=True)

# Thin ...
env['ring_thin'] = env.ring.copy(shallow=True)
line_thin = env['ring_thin']

line_thin.discard_tracker()
slicing_strategies = [
    xt.Strategy(slicing=None),  # Default
    xt.Strategy(slicing=xt.Teapot(20), element_type=xt.Bend),
    xt.Strategy(slicing=xt.Teapot(8), element_type=xt.Quadrupole),
]
line_thin.slice_thick_elements(slicing_strategies)

line_thin.build_tracker()
# line_thin.configure_radiation(model='mean')

tw_rad_thin = line_thin.twiss(eneloss_and_damping=True, strengths=True)

# Compare tunes, chromaticities, damping rates, equilibrium emittances
print('Tune comparison')
print('Thick:        ', tw_rad_thick.qx, tw_rad_thick.qy)
print('Sliced thick: ', tw_rad_sliced_thick.qx, tw_rad_sliced_thick.qy)
print('Thin:         ', tw_rad_thin.qx, tw_rad_thin.qy)

print('Chromaticity comparison')
print('Thick:        ', tw_rad_thick.dqx, tw_rad_thick.dqy)
print('Sliced thick: ', tw_rad_sliced_thick.dqx, tw_rad_sliced_thick.dqy)
print('Thin:         ', tw_rad_thin.dqx, tw_rad_thin.dqy)

print('Energy loss: ')
print('Thick:        ', tw_rad_thick.eneloss_turn)
print('Sliced thick: ', tw_rad_sliced_thick.eneloss_turn)
print('Thin:         ', tw_rad_thin.eneloss_turn)

print('Partition numbers: ')
print('Thick:        ', tw_rad_thick.partition_numbers)
print('Sliced thick: ', tw_rad_sliced_thick.partition_numbers)
print('Thin:         ', tw_rad_thin.partition_numbers)

print('Equilibrium emittances')
print('Thick:        ', tw_rad_thick.eq_gemitt_x, tw_rad_thick.eq_gemitt_y, tw_rad_thick.eq_gemitt_zeta)
print('Sliced thick: ', tw_rad_sliced_thick.eq_gemitt_x, tw_rad_sliced_thick.eq_gemitt_y, tw_rad_sliced_thick.eq_gemitt_zeta)
print('Thin:         ', tw_rad_thin.eq_gemitt_x, tw_rad_thin.eq_gemitt_y, tw_rad_thin.eq_gemitt_zeta)


from cpymad.madx import Madx
import numpy as np

import xtrack as xt

test_data_folder = '../../test_data/'
mad = Madx()

mad.call(test_data_folder + 'pimms/PIMM_orig.seq')
mad.call(test_data_folder + 'pimms/betatron.str')
mad.beam(particle='proton', gamma=1.21315778) # 200 MeV
mad.use('pimms')
seq = mad.sequence.pimms
def_expr = True

line = xt.Line.from_madx_sequence(seq, deferred_expressions=def_expr)
line.particle_ref = xt.Particles(gamma0=seq.beam.gamma,
                                 mass0=seq.beam.mass * 1e9,
                                 q0=seq.beam.charge)
line.configure_bend_model(core='full', edge='full')

line.insert_element('mysext', xt.Sextupole(length=0.2), at_s=36.)

line.insert_element(
            'septum',
            xt.LimitRect(min_x=-0.1, max_x=0.1, min_y=-0.1, max_y=0.1),
            index='pimms_start')


line.vars['k2mysext'] = 0
line.element_refs['mysext'].k2 = line.vars['k2mysext']

tw0 = line.twiss(method='4d')

line.vars['k2xrr'] = 0
opt = line.match(
    solve=False,
    method='4d',
    vary=[
        xt.VaryList(['qf1k1', 'qd1k1', 'qf2k1'], step=1e-3, tag='quad'),
        xt.VaryList(['k2xcf', 'k2xcd'], step=1e-3, tag='sext'),
    ],
    targets=[
        xt.TargetSet(qx=1.665, qy=1.72, tol=1e-6, tag='tunes'),
        xt.TargetSet(dqx=-0.1, dqy=-0.1, tol=1e-3, tag="chrom"),
        xt.Target(dx=0, tol=1e-3, at='pimms_start', tag='disp'),
    ]
)
opt.disable_targets(tag='chrom')
opt.disable_vary(tag='sext')
opt.solve()
opt.enable_all_targets()
opt.enable_all_vary()
opt.solve()

tw1 = line.twiss(method='4d')

line.vars['k2xrr'] = 8.65

tw2 = line.twiss(method='4d')


import matplotlib.pyplot as plt
plt.close('all')
plt.figure(33)
ax1 = plt.subplot(2, 1, 1)
plt.plot(tw0.s, tw0.betx, '.-')
plt.plot(tw1.s, tw1.betx, '.-')
plt.plot(tw0.s, tw0.bety, '.-')
plt.plot(tw1.s, tw1.bety, '.-')
plt.ylabel(r'$\beta$ [m]')

plt.subplot(2, 1, 2, sharex=ax1)
plt.plot(tw0.s, tw0.dx, '.-')
plt.plot(tw1.s, tw1.dx, '.-')

class ActionSeparatrix(xt.Action):

    def __init__(self, line):
        self.line = line

    def run(self):
        line = self.line
        tw = line.twiss(method='4d')

        p_test = line.build_particles(x=5e-3, px=0)
        line.track(p_test, num_turns=10000, turn_by_turn_monitor=True)
        mon_test = line.record_last_track
        norm_coord_test = tw.get_normalized_coordinates(mon_test)

        x_t = mon_test.x[:]
        px_t = mon_test.px[:]
        x_norm_t = norm_coord_test.x_norm[:]
        px_norm_t = norm_coord_test.px_norm[:]

        # Select branch closer to the x-axis
        mask_branch = (x_norm_t > 0) & (px_norm_t > -2 * x_norm_t) & (px_norm_t < 2 * x_norm_t)
        x_branch = x_t[mask_branch]
        px_branch = px_t[mask_branch]
        x_norm_branch = x_norm_t[mask_branch]
        px_norm_branch = px_norm_t[mask_branch]

        mask_fit = (x_branch > 0.01) & (x_branch < 0.02)
        poly_geom = np.polyfit(x_branch[mask_fit], px_branch[mask_fit], 1)
        poly_norm = np.polyfit(x_norm_branch[mask_fit], px_norm_branch[mask_fit], 1)

        r_sep_norm = np.abs(poly_norm[1]) / np.sqrt(poly_norm[0]**2 + 1)

        out = {
            'poly_geom': poly_geom,
            'poly_norm': poly_norm,
            'r_sep_norm': r_sep_norm,
            'slope': poly_geom[0],
        }

        return out

action_sep = ActionSeparatrix(line)
res0 = action_sep.run()

opt = line.match(
    solve=False,
    method='4d',
    vary=[
        xt.VaryList(['k2xrr', 'k2mysext'], step=1e-2, tag='resonance'),
        xt.VaryList(['k2xcf', 'k2xcd'], step=1e-3, tag='chromaticity'),
    ],
    targets=[
        action_sep.target('r_sep_norm', 1e-3, tol=1e-4, tag='resonance'),
        action_sep.target('slope', 0, tol=1e-5, tag='resonance'),
        xt.TargetSet(dqx=-4, dqy=-1, tol=1e-3, tag='chromaticity'),
    ]
)
# opt.disable_targets(tag='chromaticity')
# opt.disable_vary(tag='chromaticity')
# opt.solve()

# opt.enable_all_targets()
# opt.enable_all_vary()
# opt.solve()

res = action_sep.run()

poly_geom = res['poly_geom']
poly_norm = res['poly_norm']

x_fit_geom = np.linspace(-0.1, 0.1, 10)
px_fit_geom = poly_geom[0] * x_fit_geom + poly_geom[1]
x_fit_norm = np.linspace(-0.1, 0.1, 10)
px_fit_norm = poly_norm[0] * x_fit_norm + poly_norm[1]

tw = line.twiss(method='4d')

p = line.build_particles(
    method='4d', x=np.linspace(0, 1e-2, 20), px=0, y=0, py=0)

line.track(p, num_turns=10000, turn_by_turn_monitor=True, time=True)
mon = line.record_last_track
norm_coord = tw.get_normalized_coordinates(mon)

import matplotlib.pyplot as plt
plt.close('all')
plt.figure(100)
plt.plot(mon.x.T, mon.px.T, '.', markersize=1)
# plt.plot(x_branch, px_branch, '.k', markersize=3)
plt.plot(x_fit_geom, px_fit_geom, 'grey')
plt.ylabel(r'$p_x$')
plt.xlabel(r'$x$ [m]')

plt.figure(101)
plt.plot(norm_coord.x_norm.T, norm_coord.px_norm.T, '.', markersize=1)
# plt.plot(x_norm_branch, px_norm_branch, '.k', markersize=3)
plt.plot(x_fit_norm, px_fit_norm, 'grey')
plt.axis('equal')

plt.show()


import xtrack as xt
import xpart as xp
import xobjects as xo
from xpart.longitudinal.rf_bucket import RFBucket

import numpy as np
from scipy.constants import c as clight
from scipy.constants import e as qe
import matplotlib.pyplot as plt

gamma0 = 3 # defines the energy of the beam
gamma_transition = 4.
momentum_compaction_factor = 1 / gamma_transition**2

particle_ref = xt.Particles(gamma0=gamma0,
                            mass0=xt.PROTON_MASS_EV)

circumference = 1000.
t_rev = circumference / (particle_ref.beta0[0] * clight)
f_rev = 1 / t_rev

energy_ref_increment =  50e3 # eV

eta = momentum_compaction_factor - 1 / particle_ref.gamma0[0]**2
assert eta < 0

h_rf = 40

f_rf = h_rf * f_rev
v_rf = 100e3
lag_rf = 180. if eta > 0. else 0.

# Compute momentum increment using auxiliary particle
dp0c_eV = energy_ref_increment / particle_ref.beta0[0]

phi = np.arcsin(dp0c_eV * particle_ref.beta0[0] / v_rf)
if eta > 0:
    phi = np.pi - phi
lag_rf = np.rad2deg(phi)

otm = xt.LineSegmentMap(
    betx=1., bety=1,
    qx=6.3, qy=6.4,
    momentum_compaction_factor=momentum_compaction_factor,
    longitudinal_mode="nonlinear",
    voltage_rf=v_rf,
    frequency_rf=f_rf,
    lag_rf=lag_rf,
    length=circumference,
    energy_ref_increment=energy_ref_increment
)

line = xt.Line(elements={'otm': otm}, particle_ref=particle_ref)

tw = line.twiss()
xo.assert_allclose(tw.slip_factor, eta, atol=1e-3, rtol=0)
xo.assert_allclose(tw.qs, 0.0032729, atol=1e-7, rtol=0)

rfb = line._get_bucket()

# Mostly checking that they do not change
xo.assert_allclose(line['otm'].lag_rf[0], 30., # degrees
                   atol=1e-4, rtol=0)

xo.assert_allclose(np.rad2deg(rfb.dphi[0]), 30.,
                   atol=1e-4, rtol=0)

xo.assert_allclose(rfb.z_sfp, 0, atol=1e-4, rtol=0)
xo.assert_allclose(rfb.z_ufp, -8.33333, atol=1e-4, rtol=0)
xo.assert_allclose(rfb.z_left, rfb.z_ufp, atol=1e-4, rtol=0)
xo.assert_allclose(rfb.z_right, 4.7699056, atol=1e-4, rtol=0)

xo.assert_allclose(rfb.h_sfp(), -30.78255, atol=1e-3, rtol=0)
xo.assert_allclose(rfb.h_sfp(), rfb.hamiltonian(0, 0), atol=1e-3, rtol=0)
xo.assert_allclose(rfb.h_sfp(make_convex=True), 30.78255, atol=1e-3, rtol=0)
xo.assert_allclose(rfb.h_sfp(make_convex=True),
                   rfb.hamiltonian(rfb.z_sfp, 0, make_convex=True),
                   atol=1e-3, rtol=0)

# Build separatrix
z_separatrix_up = np.linspace(rfb.z_left, rfb.z_right, 1000)
delta_separatrix_up = rfb.separatrix(z_separatrix_up)

z_separatrix = np.array(
    list(z_separatrix_up) + list(z_separatrix_up[::-1]))
delta_separatrix = np.array(
    list(delta_separatrix_up) + list(-delta_separatrix_up[::-1]))

# Hamiltonian is defined to be zero on the separatrix
xo.assert_allclose(rfb.hamiltonian(z_separatrix, delta_separatrix), 0,
                   atol=1e-3, rtol=0)
xo.assert_allclose(rfb.hamiltonian(z_separatrix, delta_separatrix, make_convex=True), 0,
                   atol=1e-3, rtol=0)

# Check that the separatrix behaves as such in tracking
p = line.build_particles(delta=delta_separatrix[::10]*0.99, zeta=z_separatrix[::10]*0.99)
line.track(p, turn_by_turn_monitor=True, num_turns=3000)
mon = line.record_last_track
assert np.all(mon.zeta < rfb.z_right)
assert np.all(mon.zeta > rfb.z_left)

p = line.build_particles(delta=delta_separatrix[::10]*1.01, zeta=z_separatrix[::10]*1.01)
line.track(p, turn_by_turn_monitor=True, num_turns=3000)
mon = line.record_last_track
assert not np.all(mon.zeta < rfb.z_right)
assert not np.all(mon.zeta > rfb.z_left)

# Check the stable fixed point against tracking
p = line.build_particles(delta=0, zeta=rfb.z_sfp)
line.track(p, turn_by_turn_monitor=True, num_turns=3000)
mon = line.record_last_track
xo.assert_allclose(mon.zeta, rfb.z_sfp, atol=2e-3*(rfb.z_right - rfb.z_left),
                   rtol=0)
bucket_height = rfb.separatrix(rfb.z_sfp)[0]
xo.assert_allclose(mon.delta, 0, atol=2e-2*bucket_height, rtol=0)

# Fix numpy random seed
np.random.seed(0)

# Match a bunch
sigma_z = 2.
p, matcher = xp.generate_matched_gaussian_bunch(
    line=line,
    num_particles=100_000,
    nemitt_x=2.5e-6,
    nemitt_y=2.5e-6,
    sigma_z=sigma_z,
    return_matcher=True)
sigma_delta = p.delta.std()

assert np.all(p.zeta < rfb.z_right)
assert np.all(p.zeta > rfb.z_left)
assert np.all(p.delta < bucket_height)
assert np.all(p.delta > -bucket_height)

xo.assert_allclose(p.delta.max(), bucket_height, atol=0, rtol=0.03)
xo.assert_allclose(p.delta.min(), -bucket_height, atol=0, rtol=0.03)
xo.assert_allclose(p.zeta.max(), rfb.z_right, atol=0, rtol=0.03)
xo.assert_allclose(p.zeta.min(), rfb.z_left, atol=0, rtol=0.05) # this part of the bucket is poorly populated
xo.assert_allclose(p.zeta.std(), sigma_z, atol=0, rtol=0.001)

# Check that the distribution stays roughly stable over one synchrotron period
p, matcher = xp.generate_matched_gaussian_bunch(
    line=line,
    num_particles=10_000,
    nemitt_x=2.5e-6,
    nemitt_y=2.5e-6,
    sigma_z=sigma_z,
    return_matcher=True)

num_turns = int(np.round(1/tw.qs))
log_every = 3
n_log = num_turns // log_every
mon = xt.ParticlesMonitor(
    start_at_turn=0,
    stop_at_turn=1,
    n_repetitions=n_log,
    repetition_period=log_every,
    num_particles=len(p.x))

line.track(p, num_turns=num_turns, turn_by_turn_monitor=mon,
           with_progress=10)

z_mean = np.squeeze(np.mean(mon.zeta, axis=1))
z_std = np.squeeze(np.std(mon.zeta, axis=1))
delta_mean = np.squeeze(np.mean(mon.delta, axis=1))
delta_std = np.squeeze(np.std(mon.delta, axis=1))

xo.assert_allclose(z_mean, np.mean(z_mean), atol=0.02*sigma_z)
xo.assert_allclose(z_std, np.mean(z_std), atol=0.02*sigma_z)
xo.assert_allclose(delta_mean, np.mean(delta_mean), atol=0.02*sigma_delta)
xo.assert_allclose(delta_std, np.mean(delta_std), atol=0.02*sigma_delta)



import numpy as np

import numpy as np
import pandas as pd

import xtrack as xt

from xtrack.slicing import Teapot, Strategy

import matplotlib.pyplot as plt

line = xt.Line.from_json('psb_03_with_chicane_corrected.json')
line.build_tracker()

line.vars['on_chicane_k0'] = 1
line.vars['on_chicane_k2'] = 1
line.vars['on_chicane_beta_corr'] = 0
line.vars['on_chicane_tune_corr'] = 0

line_thick = line.copy()
line_thick.build_tracker()

line.discard_tracker()

slicing_strategies = [
    Strategy(slicing=Teapot(1)),  # Default
    Strategy(slicing=Teapot(2), element_type=xt.Bend),
    Strategy(slicing=Teapot(8), element_type=xt.Quadrupole),
]

print("Slicing thick elements...")
line.slice_thick_elements(slicing_strategies)

line.to_json('psb_04_with_chicane_corrected_thin.json')

line.build_tracker()

tw_thin = line.twiss()
tw_thick = line_thick.twiss()

print('\n')
print(f'Qx: thick {tw_thin.qx:.4f} thin {tw_thick.qx:.4f}, diff {tw_thin.qx-tw_thick.qx:.4e}')
print(f'Qy: thick {tw_thin.qy:.4f} thin {tw_thick.qy:.4f}, diff {tw_thin.qy-tw_thick.qy:.4e}')
print(f"Q'x: thick {tw_thin.dqx:.4f} thin {tw_thick.dqx:.4f}, diff {tw_thin.dqx-tw_thick.dqx:.4f}")
print(f"Q'y: thick {tw_thin.dqy:.4f} thin {tw_thick.dqy:.4f}, diff {tw_thin.dqy-tw_thick.dqy:.4f}")

bety_interp = np.interp(tw_thick.s, tw_thin.s, tw_thin.bety)
print(f"Max beta beat: {np.max(np.abs(tw_thick.bety/bety_interp - 1)):.4e}")

plt.close('all')

delta_values = np.linspace(-0.001, 0.001, 100)
qy_thick_list = 0 * delta_values
qx_thick_list = 0 * delta_values
qy_thin_list = 0 * delta_values
qx_thin_list = 0 * delta_values

for ii, dd in enumerate(delta_values):
    tt_thin = line.twiss(delta0=dd)
    tt_thick = line_thick.twiss(delta0=dd)

    qy_thick_list[ii] = tt_thick.qy
    qx_thick_list[ii] = tt_thick.qx
    qy_thin_list[ii] = tt_thin.qy
    qx_thin_list[ii] = tt_thin.qx

plt.close('all')

t_test = np.linspace(0, 6e-3, 100)

k0_bsw1 = []
k2_bsw1 = []
k0_bsw2 = []
k2_bsw2 = []
qx = []
qy = []
bety_at_mker = []
bety_at_mker_uncorrected = []
qy_uncorrected = []
for ii, tt in enumerate(t_test):
    print(f'Twiss at t = {tt*1e3:.2f} ms   ', end='\r', flush=True)
    line.vars['t_turn_s'] = tt

    line.vars['on_chicane_beta_corr'] = 1
    line.vars['on_chicane_tune_corr'] = 1
    tw = line.twiss()

    qx.append(tw.qx)
    qy.append(tw.qy)
    bety_at_mker.append(tw['bety', 'mker_match'])
    k0_bsw1.append(line['bi1.bsw1l1.1'].knl[0] / line['bi1.bsw1l1.1'].length)
    k2_bsw1.append(line['bi1.bsw1l1.1'].knl[2] / line['bi1.bsw1l1.1'].length)
    k0_bsw2.append(line['bi1.bsw1l1.2'].knl[0] / line['bi1.bsw1l1.2'].length)
    k2_bsw2.append(line['bi1.bsw1l1.2'].knl[2] / line['bi1.bsw1l1.2'].length)

    line.vars['on_chicane_beta_corr'] = 0
    line.vars['on_chicane_tune_corr'] = 0
    tw_uncorr = line.twiss()
    bety_at_mker_uncorrected.append(tw_uncorr['bety', 'mker_match'])
    qy_uncorrected.append(tw_uncorr.qy)

import matplotlib.pyplot as plt
plt.close('all')
plt.figure(1)
sp1 = plt.subplot(2,1,1)
plt.plot(t_test*1e3, k0_bsw1, label='k0 bsw1')
plt.plot(t_test*1e3, k0_bsw2, label='k0 bsw2')
plt.legend()
plt.subplot(2,1,2, sharex=sp1)
plt.plot(t_test*1e3, k2_bsw1, label='k2l bsw1')
plt.plot(t_test*1e3, k2_bsw2, label='k2l bsw2')
plt.legend()
plt.xlabel('time [ms]')

plt.figure(2)
sp1 = plt.subplot(2,1,1, sharex=sp1)
plt.plot(t_test*1e3, qy, label='qy')
plt.plot(t_test*1e3, qy_uncorrected, label='qy (uncorrected)')
plt.legend()
sp2 = plt.subplot(2,1,2, sharex=sp1)
plt.plot(t_test*1e3, bety_at_mker , label='bety at marker')
plt.plot(t_test*1e3, bety_at_mker_uncorrected, label='bety at marker (uncorrected)')
plt.legend()
plt.xlabel('time [ms]')


plt.figure(101)
plt.plot(delta_values, qx_thin_list, label='qy thin')
plt.plot(delta_values, qx_thick_list, label='qy thick')
plt.plot(delta_values, qy_thin_list, label='qx thin')
plt.plot(delta_values, qy_thick_list, label='qx thick')
plt.xlabel('delta')
plt.ylabel('tune')

# plot difference
plt.figure(102)
plt.plot(delta_values, qx_thick_list - qx_thin_list, label='qx thick - thin')
plt.plot(delta_values, qy_thick_list - qy_thin_list, label='qy thick - thin')
plt.xlabel('delta')
plt.ylabel('tune')
plt.legend()

import json
with open('ptc_ref.json', 'r') as fid: # generated by xtrack/examples/psb/001_compare_against_ptc.py
    ptc_ref = json.load(fid)

for kk, vv in ptc_ref.items():
    ptc_ref[kk] = np.array(vv)

t_test = ptc_ref['t_test']
qx_ptc = ptc_ref['qx_ptc']
qy_ptc = ptc_ref['qy_ptc']
dqx_ptc = ptc_ref['dqx_ptc']
dqy_ptc = ptc_ref['dqy_ptc']
bety_at_scraper_ptc = ptc_ref['bety_at_scraper_ptc']

# Check against ptc (correction off)
line.vars['on_chicane_beta_corr'] = 0
line.vars['on_chicane_tune_corr'] = 0
qx_thick = []
qy_thick = []
dqx_thick = []
dqy_thick = []
bety_at_scraper_thick = []
qx_thin = []
qy_thin = []
dqx_thin = []
dqy_thin = []
bety_at_scraper_thin = []
for ii, tt in enumerate(t_test):
    print(f'Check against ptc, twiss at t = {tt*1e3:.2f} ms   ', end='\r', flush=True)
    line_thick.vars['t_turn_s'] = tt
    line.vars['t_turn_s'] = tt

    tw_thick = line_thick.twiss()
    bety_at_scraper_thick.append(tw_thick['bety', 'br.stscrap22'])
    qx_thick.append(tw_thick.qx)
    qy_thick.append(tw_thick.qy)
    dqx_thick.append(tw_thick.dqx)
    dqy_thick.append(tw_thick.dqy)

    tw_thin = line.twiss()
    bety_at_scraper_thin.append(tw_thin['bety', 'br.stscrap22'])
    qx_thin.append(tw_thin.qx)
    qy_thin.append(tw_thin.qy)
    dqx_thin.append(tw_thin.dqx)
    dqy_thin.append(tw_thin.dqy)


assert np.allclose(qx_thick, qx_ptc, atol=2e-4, rtol=0)
assert np.allclose(qy_thick, qy_ptc, atol=2e-4, rtol=0)
assert np.allclose(qx_thin, qx_ptc, atol=1e-3, rtol=0)
assert np.allclose(qy_thin, qy_ptc, atol=1e-3, rtol=0)
assert np.allclose(dqx_thick, dqx_ptc, atol=0.5, rtol=0)
assert np.allclose(dqy_thick, dqy_ptc, atol=0.5, rtol=0)
assert np.allclose(dqx_thin, dqx_ptc, atol=0.5, rtol=0)
assert np.allclose(dqy_thin, dqy_ptc, atol=0.5, rtol=0)
assert np.allclose(bety_at_scraper_thick, bety_at_scraper_ptc, atol=0, rtol=1e-2)
assert np.allclose(bety_at_scraper_thin, bety_at_scraper_ptc, atol=0, rtol=2e-2)

# Check correction
line.vars['on_chicane_beta_corr'] = 1
line.vars['on_chicane_tune_corr'] = 1
line_thick.vars['on_chicane_beta_corr'] = 1
line_thick.vars['on_chicane_tune_corr'] = 1
qx_thick_corr = []
qy_thick_corr = []
bety_at_scraper_thick_corr = []
qx_thin_corr = []
qy_thin_corr = []
bety_at_scraper_thin_corr = []
for ii, tt in enumerate(t_test):
    print(f'Check correction, twiss at t = {tt*1e3:.2f} ms   ', end='\r', flush=True)
    line_thick.vars['t_turn_s'] = tt
    line.vars['t_turn_s'] = tt

    tw_thick = line_thick.twiss()
    bety_at_scraper_thick_corr.append(tw_thick['bety', 'br.stscrap22'])
    qx_thick_corr.append(tw_thick.qx)
    qy_thick_corr.append(tw_thick.qy)

    tw_thin = line.twiss()
    bety_at_scraper_thin_corr.append(tw_thin['bety', 'br.stscrap22'])
    qx_thin_corr.append(tw_thin.qx)
    qy_thin_corr.append(tw_thin.qy)

qx_thick_corr = np.array(qx_thick_corr)
qy_thick_corr = np.array(qy_thick_corr)
bety_at_scraper_thick_corr = np.array(bety_at_scraper_thick_corr)
qx_thin_corr = np.array(qx_thin_corr)
qy_thin_corr = np.array(qy_thin_corr)
bety_at_scraper_thin_corr = np.array(bety_at_scraper_thin_corr)

assert np.allclose(qx_thick_corr, qx_ptc[-1], atol=3e-3, rtol=0)
assert np.allclose(qy_thick_corr, qy_ptc[-1], atol=3e-3, rtol=0)
assert np.allclose(qx_thin_corr, qx_ptc[-1], atol=3e-3, rtol=0)
assert np.allclose(qy_thin_corr, qy_ptc[-1], atol=3e-3, rtol=0)
assert np.allclose(bety_at_scraper_thick_corr, bety_at_scraper_ptc[-1], atol=0, rtol=3e-2)
assert np.allclose(bety_at_scraper_thin_corr, bety_at_scraper_ptc[-1], atol=0, rtol=3e-2)

plt.show()



import xtrack as xt
import xobjects as xo
import numpy as np

env = xt.Environment()

env.new('mq', 'Quadrupole', length=1)

env.new_line(name='l1', length=20, components=[
    env.place('mq', at=10)])

env['l1'].get_table().cols['s_start s_center s_end']

tt = env['l1'].get_table()

assert np.all(tt.name == np.array(['drift_1', 'mq', 'drift_2', '_end_point']))
xo.assert_allclose(tt.s_center, np.array([ 4.75, 10.  , 15.25, 20.  ]),
                   rtol=0, atol=1e-10)


import xtrack as xt
import numpy as np

lhc = xt.Environment.from_json('mddata/step2.json')

line = lhc.b1
line.twiss_default.update({'strengths': False, 'method': '4d'})
tw0 = line.twiss()


tt = line.get_table()

observable_list = ['betx', 'bety', 'mux', 'muy', 'dx']

# obs_points = tt.rows['bpm.*'].name
import tfs
meas_betx = tfs.read_tfs('mddata/beta_phase_x.tfs')
obs_points_import = list(map(lambda ss: ss.lower(), list(meas_betx['NAME'].values)))
# Resort
obs_points = tw0.rows[obs_points_import].name

corr_elements = []
corr_vars = [nn for nn in line.vars.get_table().rows['kq.*.b1$'].name
              if 'from' not in nn]

corr_vars = corr_vars[:10] # DEBUUUUUUUG

# corr_vars = []
# corr_elements = [nn for nn in list(tt.rows[tt.element_type == 'Quadrupole'].name)
#                  if 'mqm' in nn]

correctors = []
for nn in corr_vars:
    correctors.append(('var', nn))
for nn in corr_elements:
    correctors.append(('element', nn))

response = {oo: np.zeros((len(obs_points), len(correctors)))
            for oo in observable_list}

dk = 0.5e-5
for ii, cc in enumerate(correctors):
    nn = cc[1]
    corr_type = cc[0]
    print(f'Processing {ii}/{len(correctors)}')
    if corr_type == 'var':
        line.vars[nn] += dk
    elif corr_type == 'element':
        line[nn].k1 += dk

    twp = line.twiss()

    if corr_type == 'var':
        line.vars[nn] -= dk
    elif corr_type == 'element':
        line[nn].k1 -= dk

    for observable in observable_list:
        response[observable][:, ii] = (
            twp.rows[obs_points][observable] - tw0.rows[obs_points][observable]) / dk


# line.vars['kq7.r5b1'] *= 1.01
# tw = line.twiss()
tw_meas = xt.Table({'name': np.array(obs_points_import), 'betx': meas_betx['BETX']._values})

tw = xt.Table({'name': np.array(obs_points),
               'betx': np.array([tw_meas['betx', nn] for nn in obs_points])})

from xtrack.trajectory_correction import _compute_correction

correct_on_observables = ['betx']

err = None
response_matrix = None
for cc in correct_on_observables:
    ee = tw.rows[obs_points][cc] - tw0.rows[obs_points][cc]
    rr = response[cc]

    if err is None:
        err = ee
        response_matrix = rr
    else:
        err = np.concatenate((err, ee))
        response_matrix = np.concatenate((response_matrix, rr), axis=0)

correction_svd = _compute_correction(err, response_matrix, rcond=1e-2)
correction_micado = _compute_correction(err, response_matrix, n_micado=2)

i_micado = np.argmax(np.abs(correction_micado))
print(f'MICADO correction: {correction_micado[i_micado]:.2e} at {correctors[i_micado]}')

import matplotlib.pyplot as plt
plt.close('all')
plt.figure(1)
plt.plot(correction_svd, '.', label='SVD')
plt.plot(correction_micado, '.', label='MICADO')

plt.figure(2)
plt.plot(tw0.rows[obs_points].s, tw.betx / tw0.rows[obs_points].betx -1)

plt.legend()
plt.show()


import numpy as np
import matplotlib.pyplot as plt

import xtrack as xt

# Build a simple ring
env = xt.Environment()
pi = np.pi
lbend = 3
line = env.new_line(components=[
    # Three dipoles to make a closed orbit bump
    env.new('d0.1',  xt.Drift,  length=0.05),
    env.new('bumper_0',  xt.Bend, length=0.05, k0=0, h=0),
    env.new('d0.2',  xt.Drift, length=0.3),
    env.new('bumper_1',  xt.Bend, length=0.05, k0=0, h=0),
    env.new('d0.3',  xt.Drift, length=0.3),
    env.new('bumper_2',  xt.Bend, length=0.05, k0=0, h=0),
    env.new('d0.4',  xt.Drift, length=0.2),

    # Simple ring with two FODO cells
    env.new('mqf.1', xt.Quadrupole, length=0.3, k1=0.1),
    env.new('d1.1',  xt.Drift, length=1),
    env.new('mb1.1', xt.Bend, length=lbend, k0=pi / 2 / lbend, h=pi / 2 / lbend),
    env.new('d2.1',  xt.Drift, length=1),
    env.new('mqd.1', xt.Quadrupole, length=0.3, k1=-0.7),
    env.new('d3.1',  xt.Drift, length=1),
    env.new('mb2.1', xt.Bend, length=lbend, k0=pi / 2 / lbend, h=pi / 2 / lbend),
    env.new('d3.4',  xt.Drift, length=1),
    env.new('mqf.2', xt.Quadrupole, length=0.3, k1=0.1),
    env.new('d1.2',  xt.Drift, length=1),
    env.new('mb1.2', xt.Bend, length=lbend, k0=pi / 2 / lbend, h=pi / 2 / lbend),
    env.new('d2.2',  xt.Drift, length=1),
    env.new('mqd.2', xt.Quadrupole, length=0.3, k1=-0.7),
    env.new('d3.2',  xt.Drift, length=1),
    env.new('mb2.2', xt.Bend, length=lbend, k0=pi / 2 / lbend, h=pi / 2 / lbend),
])

kin_energy_0 = 50e6 # 50 MeV
line.particle_ref = xt.Particles(energy0=kin_energy_0 + xt.PROTON_MASS_EV, # total energy
                                 mass0=xt.PROTON_MASS_EV)

# Twiss
tw = line.twiss(method='4d')

# Power the correctors to make a closed orbit bump
line.vars['bumper_strength'] = 0.
line.element_refs['bumper_0'].k0 = -line.vars['bumper_strength']
line.element_refs['bumper_1'].k0 = 2 * line.vars['bumper_strength']
line.element_refs['bumper_2'].k0 = -line.vars['bumper_strength']

# Drive the correctors with a sinusoidal function
T_sin = 100e-6
sin = line.functions.sin
line.vars['bumper_strength'] = (0.1 * sin(2 * np.pi / T_sin * line.vars['t_turn_s']))


# --- Probe behavior with twiss at different t_turn_s ---

t_test = np.linspace(0, 100e-6, 15)
tw_list = []
bumper_0_list = []
for tt in t_test:
    line.vars['t_turn_s'] = tt
    bumper_0_list.append(line.element_refs['bumper_0'].k0) # Inspect bumper
    tw_list.append(line.twiss(method='4d')) # Twiss

# Plot
plt.close('all')
plt.figure(1, figsize=(6.4*1.2, 4.8*0.85))
ax1 = plt.subplot(1, 2, 1, xlabel='Time [us]', ylabel='x at bump center [mm]')
ax2 = plt.subplot(1, 2, 2, xlabel='s [m]', ylabel='x [mm]')

colors = plt.cm.jet(np.linspace(0, 1, len(t_test)))

for ii, tt in enumerate(t_test):
    tw_tt = tw_list[ii]
    ax1.plot(tt * 1e6, tw_tt['x', 'bumper_1'] * 1e3, 'o', color=colors[ii])
    ax2.plot(tw_tt.s, tw_tt.x * 1e3, color=colors[ii], label=f'{tt * 1e6:.1f} us')

ax2.set_xlim(0, tw_tt['s', 'bumper_2'] + 1)
plt.subplots_adjust(left=.1, right=.97, top=.92, wspace=.27)
ax2.legend()


# --- Track particles with time-dependent bumpers ---

num_particles = 100
num_turns = 1000

# Install monitor in the middle of the bump
monitor = xt.ParticlesMonitor(num_particles=num_particles,
                              start_at_turn=0,
                              stop_at_turn =num_turns)
line.discard_tracker()
line.insert('monitor', monitor, at='bumper_1@start')
line.build_tracker()

# Generate particles
particles = line.build_particles(
                        x=np.random.uniform(-0.1e-3, 0.1e-3, num_particles),
                        px=np.random.uniform(-0.1e-3, 0.1e-3, num_particles))

# Enable time-dependent variables
line.enable_time_dependent_vars = True

# Track
line.track(particles, num_turns=num_turns, with_progress=True)

# Plot
plt.figure(2, figsize=(6.4*0.8, 4.8*0.8))
plt.plot(1e6 * monitor.at_turn.T * tw.T_rev0, 1e3 * monitor.x.T,
         lw=1, color='r', alpha=0.05)
plt.plot(1e6 * monitor.at_turn[0, :] * tw.T_rev0, 1e3 * monitor.x.mean(axis=0),
         lw=2, color='k')
plt.xlabel('Time [us]')
plt.ylabel('x [mm]')
plt.show()



import xtrack as xt

d_zeta = 1e-3

collider = xt.Environment.from_json(
    '../../test_data/hllhc15_collider/collider_00_from_mad.json'
)
collider.build_trackers()

collider.vars['vrf400'] = 16
collider.vars['on_crab1'] = -190
collider.vars['on_crab5'] = -190

line = collider.lhcb1

tw_plus_closed = line.twiss(
    method='4d', zeta0=d_zeta, delta0=0, freeze_longitudinal=True)
tw_minus_closed = line.twiss(
    method='4d', zeta0=-d_zeta, delta0=0, freeze_longitudinal=True)
cl_dx_zeta_4d_rf_on_crab_off = (tw_plus_closed.x - tw_minus_closed.x)/(tw_plus_closed.zeta - tw_minus_closed.zeta)

import matplotlib.pyplot as plt
plt.close('all')
plt.figure(1)
plt.plot(tw_plus_closed.s, cl_dx_zeta_4d_rf_on_crab_off, label='rf on, crab on', color='g')

tw = line.twiss()
W = tw.W_matrix

# copilot please implement the following equation
# \begin{equation}
# D_x^\zeta = \left(W_{15} -\frac{W_{16}W_{65}}{W_{66}}\right)\left( W_{55}\ -  \frac{W_{56}W_{65}}{W_{66}} \right)^{-1}
# \end{equation}

dx_zeta = (W[:, 0, 4] - W[:, 0, 5] * W[:, 5, 4] / W[:, 5, 5]) / (
             W[:, 4, 4] - W[:, 4, 5] * W[:, 5, 4] / W[:, 5, 5])

plt.show()


import pathlib
import json

import numpy as np
from cpymad.madx import Madx

import xpart as xp
import xtrack as xt
import xobjects as xo
from xobjects.test_helpers import for_all_test_contexts

test_data_folder = pathlib.Path(
    __file__).parent.joinpath('../test_data').absolute()

@for_all_test_contexts
def test_orbit_correction_basics(test_context):

    line = xt.Line.from_json(test_data_folder
                             / 'hllhc15_thick/lhc_thick_with_knobs.json')
    line.build_tracker(_context=test_context)
    tt = line.get_table()

    # Define elements to be used as monitors for orbit correction
    # (for LHC all element names starting by "bpm" and not ending by "_entry" or "_exit")
    #tt_monitors = tt.rows['bpm.*'].rows['.*(?<!_entry)$'].rows['.*(?<!_exit)$']
    tt_monitors = tt.rows['bpm.*','.*(?<!_entry)$','.*(?<!_exit)$']
    line.steering_monitors_x = tt_monitors.name
    line.steering_monitors_y = tt_monitors.name

    # Define elements to be used as correctors for orbit correction
    # (for LHC all element namesstarting by "mcb.", containing "h." or "v.")
    tt_h_correctors = tt.rows['mcb.*'].rows[r'.*h\..*']
    line.steering_correctors_x = tt_h_correctors.name
    tt_v_correctors = tt.rows['mcb.*'].rows[r'.*v\..*']
    line.steering_correctors_y = tt_v_correctors.name

    # Reference twiss (no misalignments)
    tw_ref = line.twiss4d()

    # Introduce misalignments on all quadrupoles
    tt = line.get_table()
    tt_quad = tt.rows[tt.element_type == 'Quadrupole']
    rgen = np.random.RandomState(1) # fix seed for random number generator
                                    # (to have reproducible results)
    shift_x = rgen.randn(len(tt_quad)) * 0.01e-3 # 0.01 mm rms shift on all quads
    shift_y = rgen.randn(len(tt_quad)) * 0.01e-3 # 0.01 mm rms shift on all quads
    for nn_quad, sx, sy in zip(tt_quad.name, shift_x, shift_y):
        line.element_refs[nn_quad].shift_x = sx
        line.element_refs[nn_quad].shift_y = sy

    # Twiss before correction
    tw_before = line.twiss4d()

    # Correct orbit
    orbit_correction = line.correct_trajectory(twiss_table=tw_ref)

    # Twiss after correction
    tw_after = line.twiss4d()

    # Extract correction strength
    kicks_x = orbit_correction.x_correction.get_kick_values()
    kicks_y = orbit_correction.y_correction.get_kick_values()

    assert tw_before.x.std() > 1e-3
    assert tw_before.y.std() > 1e-3
    assert tw_after.y.std() < 5e-6
    assert tw_after.x.std() < 5e-6

    assert kicks_x.std() < 5e-6
    assert kicks_y.std() < 5e-6

    xo.assert_allclose(np.abs(kicks_x).max(), 17e-6, atol=2e-6)
    xo.assert_allclose(np.abs(kicks_y).max(), 5e-6, atol=1e-6)

@for_all_test_contexts
def test_orbit_correction_micado(test_context):

    line = xt.Line.from_json(test_data_folder
        / 'hllhc15_thick/lhc_thick_with_knobs.json')
    line.build_tracker(_context=test_context)
    tt = line.get_table()

    # Define elements to be used as monitors for orbit correction
    # (for LHC all element names starting by "bpm" and not ending by "_entry" or "_exit")
    tt_monitors = tt.rows['bpm.*'].rows['.*(?<!_entry)$'].rows['.*(?<!_exit)$']
    line.steering_monitors_x = tt_monitors.name
    line.steering_monitors_y = tt_monitors.name

    # Define elements to be used as correctors for orbit correction
    # (for LHC all element namesstarting by "mcb.", containing "h." or "v.")
    tt_h_correctors = tt.rows['mcb.*'].rows[r'.*h\..*']
    line.steering_correctors_x = tt_h_correctors.name
    tt_v_correctors = tt.rows['mcb.*'].rows[r'.*v\..*']
    line.steering_correctors_y = tt_v_correctors.name

    # Reference twiss (no misalignments)
    tw_ref = line.twiss4d()

    # Introduce misalignments on all quadrupoles
    tt = line.get_table()
    tt_quad = tt.rows[tt.element_type == 'Quadrupole']
    rgen = np.random.RandomState(1) # fix seed for random number generator
    shift_x = rgen.randn(len(tt_quad)) * 0.01e-3 # 0.01 mm rms shift on all quads
    shift_y = rgen.randn(len(tt_quad)) * 0.01e-3 # 0.01 mm rms shift on all quads
    for nn_quad, sx, sy in zip(tt_quad.name, shift_x, shift_y):
        line.element_refs[nn_quad].shift_x = sx
        line.element_refs[nn_quad].shift_y = sy

    # Twiss before correction
    tw_before = line.twiss4d()

    # Correct orbit using 5 correctors in each plane
    orbit_correction = line.correct_trajectory(twiss_table=tw_ref, n_micado=5,
                                               n_iter=1)

    # Twiss after correction
    tw_after = line.twiss4d()

    # Extract correction strength
    s_x_correctors = orbit_correction.x_correction.s_correctors
    s_y_correctors = orbit_correction.y_correction.s_correctors
    kicks_x = orbit_correction.x_correction.get_kick_values()
    kicks_y = orbit_correction.y_correction.get_kick_values()

    assert tw_before.x.std() > 1e-3
    assert tw_before.y.std() > 1e-3
    assert tw_after.y.std() < 2e-4
    assert tw_after.x.std() < 2e-4

    assert np.sum(np.abs(kicks_x)>0) == 5
    assert np.sum(np.abs(kicks_y)>0) == 5

@for_all_test_contexts
def test_orbit_correction_customize(test_context):

    line = xt.Line.from_json(test_data_folder
         / 'hllhc15_thick/lhc_thick_with_knobs.json')
    line.build_tracker(_context=test_context)
    tt = line.get_table()

    # Define elements to be used as monitors for orbit correction
    # (for LHC all element names starting by "bpm" and not ending by "_entry" or "_exit")
    tt_monitors = tt.rows['bpm.*'].rows['.*(?<!_entry)$'].rows['.*(?<!_exit)$']
    line.steering_monitors_x = tt_monitors.name
    line.steering_monitors_y = tt_monitors.name

    # Define elements to be used as correctors for orbit correction
    # (for LHC all element namesstarting by "mcb.", containing "h." or "v.")
    tt_h_correctors = tt.rows['mcb.*'].rows[r'.*h\..*']
    line.steering_correctors_x = tt_h_correctors.name
    tt_v_correctors = tt.rows['mcb.*'].rows[r'.*v\..*']
    line.steering_correctors_y = tt_v_correctors.name

    # Reference twiss (no misalignments)
    tw_ref = line.twiss4d()

    # Introduce misalignments on all quadrupoles
    tt = line.get_table()
    tt_quad = tt.rows[tt.element_type == 'Quadrupole']
    rgen = np.random.RandomState(1) # fix seed for random number generator
    shift_x = rgen.randn(len(tt_quad)) * 0.01e-3 # 0.01 mm rms shift on all quads
    shift_y = rgen.randn(len(tt_quad)) * 0.01e-3 # 0.01 mm rms shift on all quads
    for nn_quad, sx, sy in zip(tt_quad.name, shift_x, shift_y):
        line.element_refs[nn_quad].shift_x = sx
        line.element_refs[nn_quad].shift_y = sy

    # Create orbit correction object without running the correction
    orbit_correction = line.correct_trajectory(twiss_table=tw_ref, run=False)

    # Twiss before correction
    tw_before = line.twiss4d()

    # Correct
    orbit_correction.correct()

    # Remove applied correction
    orbit_correction.clear_correction_knobs()

    # Correct with a customized number of singular values
    orbit_correction.correct(n_singular_values=(200, 210))

    # Twiss after correction
    tw_after = line.twiss4d()

    # Extract correction strength
    s_x_correctors = orbit_correction.x_correction.s_correctors
    s_y_correctors = orbit_correction.y_correction.s_correctors
    kicks_x = orbit_correction.x_correction.get_kick_values()
    kicks_y = orbit_correction.y_correction.get_kick_values()

    assert tw_before.x.std() > 1e-3
    assert tw_before.y.std() > 1e-3
    assert tw_after.y.std() < 10e-6
    assert tw_after.x.std() < 10e-6

    assert kicks_x.std() < 5e-7
    assert kicks_y.std() < 5e-7

    xo.assert_allclose(np.abs(kicks_x).max(), 2e-6, atol=2e-6)
    xo.assert_allclose(np.abs(kicks_y).max(), 2e-6, atol=1e-6)

@for_all_test_contexts
def test_orbit_correction_thread(test_context):

    line = xt.Line.from_json(test_data_folder
        / 'hllhc15_thick/lhc_thick_with_knobs.json')
    line.build_tracker(_context=test_context)
    tt = line.get_table()

    # Define elements to be used as monitors for orbit correction
    # (for LHC all element names starting by "bpm" and not ending by "_entry" or "_exit")
    tt_monitors = tt.rows['bpm.*'].rows['.*(?<!_entry)$'].rows['.*(?<!_exit)$']
    line.steering_monitors_x = tt_monitors.name
    line.steering_monitors_y = tt_monitors.name

    # Define elements to be used as correctors for orbit correction
    # (for LHC all element namesstarting by "mcb.", containing "h." or "v.")
    tt_h_correctors = tt.rows['mcb.*'].rows[r'.*h\..*']
    line.steering_correctors_x = tt_h_correctors.name
    tt_v_correctors = tt.rows['mcb.*'].rows[r'.*v\..*']
    line.steering_correctors_y = tt_v_correctors.name

    # Reference twiss (no misalignments)
    tw_ref = line.twiss4d()

    # Introduce misalignments on all quadrupoles
    tt = line.get_table()
    tt_quad = tt.rows[tt.element_type == 'Quadrupole']
    rgen = np.random.RandomState(2) # fix seed for random number generator
    shift_x = rgen.randn(len(tt_quad)) * 1e-3 # 1. mm rms shift on all quads
    shift_y = rgen.randn(len(tt_quad)) * 1e-3 # 1. mm rms shift on all quads
    for nn_quad, sx, sy in zip(tt_quad.name, shift_x, shift_y):
        line.element_refs[nn_quad].shift_x = sx
        line.element_refs[nn_quad].shift_y = sy

    # Closed twiss fails (closed orbit is not found)
    # line.twiss4d()

    # Create orbit correction object without running the correction
    orbit_correction = line.correct_trajectory(twiss_table=tw_ref, run=False)

    # Thread
    threader = orbit_correction.thread(ds_thread=500., # correct in sections of 500 m
                                    rcond_short=1e-3, rcond_long=1e-3)

    kicks_x_thread = orbit_correction.x_correction.get_kick_values()
    kicks_x_thread = orbit_correction.x_correction.get_kick_values()

    # Closed twiss after threading (closed orbit is found)
    tw_after_thread = line.twiss4d()

    # Correct (with custom number of singular values)
    orbit_correction.correct(n_singular_values=200)

    # Twiss after closed orbit correction
    tw_after = line.twiss4d()

    # Extract correction strength
    s_x_correctors = orbit_correction.x_correction.s_correctors
    s_y_correctors = orbit_correction.y_correction.s_correctors
    kicks_x = orbit_correction.x_correction.get_kick_values()
    kicks_y = orbit_correction.y_correction.get_kick_values()

    assert tw_after_thread.x.std() < 2e-3
    assert tw_after_thread.y.std() < 2e-3

    assert tw_after.x.std() < 5e-4
    assert tw_after.y.std() < 5e-4

    assert kicks_x.std() < 5e-5
    assert kicks_y.std() < 5e-5

@for_all_test_contexts
def test_correct_trajectory_transfer_line(test_context):

    mad_ti2 = Madx()
    mad_ti2.call(str(test_data_folder / 'sps_to_lhc_ti2/ti2.seq'))
    mad_ti2.call(str(test_data_folder / 'sps_to_lhc_ti2/ti2_liu.str'))
    mad_ti2.beam()
    mad_ti2.use('ti2')

    line = xt.Line.from_madx_sequence(mad_ti2.sequence['ti2'])
    line.particle_ref = xt.Particles(p0c=450e9, mass0=xt.PROTON_MASS_EV, q0=1)
    line.build_tracker(_context=test_context)
    tt = line.get_table()

    # Define elements to be used as monitors for orbit correction
    # (in this case all element names starting by "bpm" and not ending by "_entry" or "_exit")
    tt_monitors = tt.rows['bpm.*'].rows['.*(?<!_entry)$'].rows['.*(?<!_exit)$']
    line.steering_monitors_x = tt_monitors.name
    line.steering_monitors_y = tt_monitors.name

    # Define elements to be used as correctors for orbit correction
    # (in this case all element names starting by "mci.", containing "h." or "v.")
    tt_h_correctors = tt.rows['mci.*'].rows[r'.*h\..*']
    line.steering_correctors_x = tt_h_correctors.name
    tt_v_correctors = tt.rows['mci.*'].rows[r'.*v\..*']
    line.steering_correctors_y = tt_v_correctors.name

    # Initial conditions from upstream ring
    init = xt.TwissInit(betx=27.77906807, bety=120.39920690,
                        alfx=0.63611880, alfy=-2.70621900,
                        dx=-0.59866300, dpx=0.01603536)

    # Reference twiss (no misalignments)
    tw_ref = line.twiss4d(start='ti2$start', end='ti2$end', init=init)

    # Introduce misalignments on all quadrupoles
    tt = line.get_table()
    tt_quad = tt.rows['mqi.*']
    rgen = np.random.RandomState(2) # fix seed for random number generator
    shift_x = rgen.randn(len(tt_quad)) * 0.1e-3 # 0.1 mm rms shift on all quads
    shift_y = rgen.randn(len(tt_quad)) * 0.1e-3 # 0.1 mm rms shift on all quads
    for nn_quad, sx, sy in zip(tt_quad.name, shift_x, shift_y):
        line.element_refs[nn_quad].shift_x = sx
        line.element_refs[nn_quad].shift_y = sy

    # Twiss before correction
    tw_before = line.twiss4d(start='ti2$start', end='ti2$end', init=init)

    # Correct trajectory
    correction = line.correct_trajectory(twiss_table=tw_ref, start='ti2$start', end='ti2$end')

    # Twiss after correction
    tw_after = line.twiss4d(start='ti2$start', end='ti2$end', init=init)

    # Extract correction strength
    kicks_x = correction.x_correction.get_kick_values()
    kicks_y = correction.y_correction.get_kick_values()

    assert tw_before.x.std() > 1.5e-3
    assert tw_before.y.std() > 0.5e-3

    assert tw_after.x.std() < 0.3e-3
    assert tw_after.y.std() < 0.3e-3

    assert kicks_x.std() < 2e-5
    assert kicks_y.std() < 2e-5

def test_orbit_correction_and_threading_shift_monitors():

    env = xt.Environment()
    env.particle_ref = xt.Particles(p0c=450e9, mass0=xt.PROTON_MASS_EV, q0=1)
    env.vars.default_to_zero = True

    env.new('mq', 'Quadrupole', length=0.8)
    env.new('bpm', 'Marker')
    env.new('corrector', 'Multipole', knl=[0])

    line = env.new_line(components=[

        env.new('line.start', 'Marker'),
        env.new('line.end', 'Marker', at=12.),

        env.new('mq1', 'mq', k1=0.2, at=3.),
        env.new('mq2', 'mq', k1=-0.2, at=5.),
        env.new('mq3', 'mq', k1=0.2, at=7.),
        env.new('mq4', 'mq', k1=-0.2, at=9.),

        env.new('bpm.s.1', 'bpm', at=0.1),
        env.new('bpm.s.2', 'bpm', at=0.5),
        env.new('bpm.q.1', 'bpm', at='mq1@start'),
        env.new('bpm.q.2', 'bpm', at='mq2@start'),
        env.new('bpm.q.3', 'bpm', at='mq3@start'),
        env.new('bpm.q.4', 'bpm', at='mq4@start'),
        env.new('bpm.e.1', 'bpm', at=11.5),
        env.new('bpm.e.2', 'bpm', at=11.9),

        env.new('corr1', 'corrector', at=1., knl=['k0l_corr1'], ksl=['k0sl_corr1']),
        env.new('corr2', 'corrector', at=2., knl=['k0l_corr2'], ksl=['k0sl_corr2']),
        env.new('corr3', 'corrector', at=10., knl=['k0l_corr3'], ksl=['k0sl_corr3']),
        env.new('corr4', 'corrector', at=11., knl=['k0l_corr4'], ksl=['k0sl_corr4']),
    ])

    # Define monitors and correctors for orbit steering
    line.steering_monitors_x = ['bpm.s.1', 'bpm.s.2',
                                'bpm.q.1', 'bpm.q.2', 'bpm.q.3', 'bpm.q.4',
                                'bpm.e.1', 'bpm.e.2']
    line.steering_correctors_x = ['corr1', 'corr2', 'corr3', 'corr4']
    line.steering_monitors_y = line.steering_monitors_x
    line.steering_correctors_y = line.steering_correctors_x

    # Twiss without misalignments
    tw0 = line.twiss4d()

    # Misalign all quadrupoles
    env.set(['mq1', 'mq2', 'mq3', 'mq4'], shift_x=1e-3, shift_y=2e-3, rot_s_rad=0.1)

    # Define BPM alignment
    bpm_alignment ={
        'bpm.q.1': {'shift_x': 1e-3, 'shift_y': 2e-3, 'rot_s_rad': 0.1},
        'bpm.q.2': {'shift_x': 1e-3, 'shift_y': 2e-3, 'rot_s_rad': 0.1},
        'bpm.q.3': {'shift_x': 1e-3, 'shift_y': 2e-3, 'rot_s_rad': 0.1},
        'bpm.q.4': {'shift_x': 1e-3, 'shift_y': 2e-3, 'rot_s_rad': 0.1},
    }

    # Correct orbit taking into account BPM alignment (centers the beam in all quadrupoles)
    correction = line.correct_trajectory(twiss_table=tw0,
                                        monitor_alignment=bpm_alignment, # <--BPM alignment
                                        run=False)
    correction.correct()

    #!end-doc-part

    tw_corr = line.twiss4d()

    xo.assert_allclose(correction.x_correction.shift_x_monitors,
                    [0., 0., 0.001, 0.001, 0.001, 0.001, 0., 0.], rtol=0, atol=1e-14)
    xo.assert_allclose(correction.x_correction.shift_y_monitors,
                    [0., 0., 0.002, 0.002, 0.002, 0.002, 0., 0.], rtol=0, atol=1e-14)
    xo.assert_allclose(correction.y_correction.shift_x_monitors,
                    [0., 0., 0.001, 0.001, 0.001, 0.001, 0., 0.], rtol=0, atol=1e-14)
    xo.assert_allclose(correction.y_correction.shift_y_monitors,
                    [0., 0., 0.002, 0.002, 0.002, 0.002, 0., 0.], rtol=0, atol=1e-14)

    # Data from previous step can be found in:
    correction.correct() # Some more steps to log the position
    xo.assert_allclose(correction.x_correction._position_before,0, rtol=0, atol=1e-10)
    xo.assert_allclose(correction.y_correction._position_before,0, rtol=0, atol=1e-10)

    correction.clear_correction_knobs()

    for nn in ['corr1', 'corr2', 'corr3', 'corr4']:
        assert line[nn].knl[0] == 0

    correction.thread(ds_thread = 10)
    tw_thread = line.twiss4d()

    xo.assert_allclose(correction.x_correction.shift_x_monitors,
                    [0., 0., 0.001, 0.001, 0.001, 0.001, 0., 0.], rtol=0, atol=1e-14)
    xo.assert_allclose(correction.x_correction.shift_y_monitors,
                    [0., 0., 0.002, 0.002, 0.002, 0.002, 0., 0.], rtol=0, atol=1e-14)
    xo.assert_allclose(correction.y_correction.shift_x_monitors,
                    [0., 0., 0.001, 0.001, 0.001, 0.001, 0., 0.], rtol=0, atol=1e-14)
    xo.assert_allclose(correction.y_correction.shift_y_monitors,
                    [0., 0., 0.002, 0.002, 0.002, 0.002, 0., 0.], rtol=0, atol=1e-14)

    # Data from previous step can be found in:
    correction.correct() # Some more steps to log the position
    xo.assert_allclose(correction.x_correction._position_before,0, rtol=0, atol=1e-9)
    xo.assert_allclose(correction.y_correction._position_before,0, rtol=0, atol=1e-9)

    xo.assert_allclose(tw_corr.rows['mq1':'corr3'].x, 1e-3, rtol=0, atol=1e-9)
    xo.assert_allclose(tw_corr.rows['mq1':'corr3'].y, 2e-3, rtol=0, atol=1e-9)
    xo.assert_allclose(tw_thread.rows['mq1':'corr3'].x, 1e-3, rtol=0, atol=1e-4)
    xo.assert_allclose(tw_thread.rows['mq1':'corr3'].y, 2e-3, rtol=0, atol=1e-4)

    xo.assert_allclose(tw_corr.x[0], 0, rtol=0, atol=1e-9)
    xo.assert_allclose(tw_corr.y[0], 0, rtol=0, atol=1e-9)
    xo.assert_allclose(tw_thread.x[0], 0, rtol=0, atol=1e-4)
    xo.assert_allclose(tw_thread.y[0], 0, rtol=0, atol=1e-4)

def test_orbit_correction_tilt_monitors():

    env = xt.Environment()
    env.particle_ref = xt.Particles(p0c=450e9, mass0=xt.PROTON_MASS_EV, q0=1)
    env.vars.default_to_zero = True

    env.new('mq', 'Quadrupole', length=0.8)
    env.new('bpm', 'Marker')
    env.new('corrector', 'Multipole', knl=[0])

    line = env.new_line(components=[

        env.new('line.start', 'Marker'),
        env.new('line.end', 'Marker', at=12.),

        env.new('mq1', 'mq', k1='kq1', at=3.),
        env.new('mq2', 'mq', k1='kq2', at=5.),
        env.new('mq3', 'mq', k1='kq3', at=7.),
        env.new('mq4', 'mq', k1='kq4', at=9.),

        env.new('bpm1', 'bpm', at='mq1@start'),
        env.new('bpm2', 'bpm', at='mq2@start'),
        env.new('bpm3', 'bpm', at='mq3@start'),
        env.new('bpm4', 'bpm', at='mq4@start'),

        env.new('corrector1', 'corrector', at='mq1@start'),
        env.new('corrector2', 'corrector', at='mq2@start'),
        env.new('corrector3', 'corrector', at='mq3@start'),
        env.new('corrector4', 'corrector', at='mq4@start'),

        env.new('bumper1', 'corrector', at=1., knl=['k0l_bumper1'], ksl=['k0sl_bumper1']),
        env.new('bumper2', 'corrector', at=2., knl=['k0l_bumper2'], ksl=['k0sl_bumper2']),
    ])

    env.set(['mq1', 'mq2', 'mq3', 'mq4'], shift_x=1e-3, shift_y=1.5e-3,
            rot_s_rad=np.deg2rad(30.))

    # Steer to enter at the center of the first quad
    line.match(
        betx=1., bety=1.,
        vary=xt.VaryList(['k0l_bumper1', 'k0l_bumper2', 'k0sl_bumper1', 'k0sl_bumper2'],
                        step=1e-6),
        targets=xt.TargetSet(x=1e-3, px=0, y=1.5e-3, py=0, at='bpm1'),
    )

    tt = line.get_table()
    tt_quad = tt.rows['mq.*']

    env['kq1'] = 0.02
    env['kq2'] = -0.02
    env['kq3'] = 0.02
    env['kq4'] = -0.02

    dx = 2e-3

    bpm_alignment ={
        'bpm1': {'shift_x': 1e-3, 'shift_y': 1.5e-3, 'rot_s_rad': np.deg2rad(30.)},
        'bpm2': {'shift_x': 1e-3, 'shift_y': 1.5e-3, 'rot_s_rad': np.deg2rad(30.)},
        'bpm3': {'shift_x': 1e-3, 'shift_y': 1.5e-3, 'rot_s_rad': np.deg2rad(30.)},
        'bpm4': {'shift_x': 1e-3, 'shift_y': 1.5e-3, 'rot_s_rad': np.deg2rad(30.)},
    }

    tw0 = line.twiss(betx=100, bety=80)

    line.steering_monitors_x = ['bpm1', 'bpm2', 'bpm3', 'bpm4']
    line.steering_monitors_y = ['bpm1', 'bpm2', 'bpm3', 'bpm4']
    line.steering_correctors_x = ['corrector1', 'corrector3', ]
    line.steering_correctors_y = ['corrector2', 'corrector4']

    tw = line.twiss(betx=1, bety=1, x=1e-3+dx*np.cos(np.deg2rad(30)), y=1.5e-3+dx*np.sin(np.deg2rad(30)))


    correction = line.correct_trajectory(twiss_table=tw0,
                                        start='line.start', end='line.end',
                                        monitor_alignment=bpm_alignment,
                                        run=False)

    correction.correct(n_iter=1)

    # Check that there is no vertical reading in the tilted bpm BPMs
    xo.assert_allclose(correction.y_correction._position_before, 0, atol=1e-15, rtol=0)



from cpymad.madx import Madx

mad = Madx()

mad.input('''
set, format="19.15f";
option,update_from_parent=true;

radiationandrf = 1;

call,file="fccee_t.seq";

pbeam =   182.5;
exbeam = 1.46e-9;
eybeam = 2.9e-12;
nbun =    48;
npar =   2.3e11;
halfcrossingangle = 0.015;
ebeam = sqrt( pbeam^2 + emass^2 );
beam, particle=positron, npart=npar, kbunch=nbun, energy=ebeam, radiate=false, bv=+1, ex=exbeam, ey=eybeam;

use, sequence = fccee_p_ring;

voltca1save = voltca1;
voltca2save = voltca2;


!rf and radiation off (for 4d twiss)
voltca1 = 0;
voltca2 = 0;
''')

tw_thick = mad.twiss().dframe()

mad.input('''
select, flag=makethin, class=rfcavity, slice = 1, thick = false;
select, flag=makethin, class=rbend, slice = 4, thick = false;
select, flag=makethin, class=quadrupole, slice = 4, thick = false;
select, flag=makethin, class=sextupole, slice = 4;
select, flag=makethin, pattern="qc*", slice=20;
select, flag=makethin, pattern="sy*", slice=20;

makethin, sequence=fccee_p_ring, style=teapot, makedipedge=false;
use, sequence = fccee_p_ring;
''')

tw_thin = mad.twiss().dframe()

# Restore rf_voltage
mad.input('''
    voltca1 = voltca1save;
    voltca2 = voltca2save;
'''
)

import xtrack as xt
import xpart as xp
line = xt.Line.from_madx_sequence(mad.sequence.fccee_p_ring)
line.particle_ref = xp.Particles(mass0=xp.ELECTRON_MASS_EV,
                                gamma0=mad.sequence.fccee_p_ring.beam.gamma)

import json
import xobjects as xo
with open('line_no_radiation.json', 'w') as f:
    json.dump(line.to_dict(), f, cls=xo.JEncoder)


import xtrack as xt
import xobjects as xo
import numpy as np

env = xt.Environment()

line = env.new_line(
    components=[
        env.new('q0', 'Quadrupole', length=2.0, at=20.0),
        env.new('ql', 'Quadrupole', length=2.0, at=10.0),
        env.new('qr', 'Quadrupole', length=2.0, at=30),
        env.new('end', 'Marker', at=50.),
    ])

class MyElement:
    def __init__(self, myparameter):
        self.myparameter = myparameter

    def track(self, particles):
        particles.px += self.myparameter

myelem = MyElement(0.1)

line.insert(
    env.place('myname', myelem, at='qr@end'))

tt = line.get_table()
tt.show(cols=['name', 's_start', 's_end', 's_center'])

assert np.all(tt.name == np.array(
    ['drift_1', 'ql', 'drift_2', 'q0', 'drift_3', 'qr', 'myname',
     'drift_4', 'end', '_end_point']))
assert np.all(tt.element_type == np.array(
    ['Drift', 'Quadrupole', 'Drift', 'Quadrupole', 'Drift',
       'Quadrupole', 'MyElement', 'Drift', 'Marker', '']))
xo.assert_allclose(tt.s, np.array(
    [ 0.,  9., 11., 19., 21., 29., 31., 31., 50., 50.]),
    rtol=0., atol=1e-14)


import numpy as np
import matplotlib.pyplot as plt

import xtrack as xt
import xobjects as xo
import xpart as xp

import RF_Track as RFT

#############################################
##########    Input parameters    ###########
#############################################

# Bunch parameters
q0 = -1 # electrons
P0c = 100e6 # reference momentum, eV/c

# Load the 1D field map of the CLIC RF structure
T = np.loadtxt(fname='400_Ez_on-axis_TD26__vg1p8_r05_CC_EDMS.txt', delimiter='\t')
Ez = np.complex64(T[:,1] + 1.j*T[:,2]) # V/m, take columns 1 and 2 to make a complex number
hz = T[1,0] - T[0,0] # mm, mesh step
freq = 11.994e9 # Hz, RF frequency

#############################################
#######  RF-Track's part starts here  #######
#############################################

# Setup the element
RF = RFT.RF_FieldMap_1d(Ez, hz/1e3, -1, freq, +1);
RF.set_P_map(4) # W, field map input power
RF.set_P_actual(10e6) # W, actual input power
RF.set_odeint_algorithm('rk2')
RF.set_smooth(10)

#############################################
#######  RF-Track's part ends here    #######
#############################################

# Back to Xsuite

elements = {
    'sol': xt.RFT_Element(element=RF),
}

# Build the line
line = xt.Line(elements=elements, element_names=list(elements.keys()))
line.particle_ref = xt.Particles(p0c=P0c, mass0=xt.ELECTRON_MASS_EV, q0=q0)

# Define the beam using
sigma_x = 0.0001 # m, sigma x
sigma_y = 0.0001 # m, sigma y
sigma_z = 0.0005 # m, sigma z

## Build particle object on context
n_part = 10000

rng = np.random.default_rng(12345)

# track
## Choose a context
context = xo.ContextCpu()         # For CPU
# context = xo.ContextCupy()      # For CUDA GPUs
# context = xo.ContextPyopencl()  # For OpenCL GPUs

## Transfer lattice on context and compile tracking code
line.build_tracker(_context=context)

particles = xp.Particles(
    _context=context,
    x=rng.normal(scale=sigma_x, size=n_part),
    y=rng.normal(scale=sigma_y, size=n_part),
    zeta=rng.normal(scale=sigma_z, size=n_part),
    p0c=P0c, mass0=xt.ELECTRON_MASS_EV, q0=q0)

#############################################
#########   Xsuite tracking    ##############
#############################################

print('tracking starts')
line.track(particles)
print('tracking ends')

print('initial reference momentum = ', P0c/1e6, 'MeV/c')
print('final reference momentum = ', particles.p0c[0]/1e6, 'MeV/c')

#############################################
## Retrieve Twiss from RF-Track and plot  ###
#############################################

# Plot the transverse phase space
plt.figure(1)
plt.scatter(particles.x*1e3, particles.y*1e3, s=10, facecolors='none', edgecolors='b')
plt.xlabel("$x$ [mm]")
plt.ylabel("$y$ [mm]")
plt.show()

# Plot the longitudinal phase space
plt.figure(2)
plt.scatter(particles.zeta*1e3, particles.delta*1e3, s=10, facecolors='none', edgecolors='b')
plt.xlabel("$z$ [mm]")
plt.ylabel("$\delta$ [permil]")
plt.show()

# Plot the field along the axis
Z = np.linspace(0.0, RF.get_length(), 1001) # m
O = np.zeros(1001)
[E,B] = RF.get_field(O, O, Z*1e3, O) # X [mm], Y [mm], Z [mm], T [mm/c]
plt.figure(3)
plt.plot(Z.transpose(), E[:,2]/1e6)
plt.xlabel("$S$ [m]")
plt.ylabel("$E_z$ [MV/m]")
plt.show()



import json
import pathlib

import numpy as np

import xobjects as xo
import xtrack as xt

test_data_folder = pathlib.Path(
        __file__).parent.joinpath('../test_data').absolute()

def test_tapering_and_twiss_with_radiation():

    filename = test_data_folder / 'clic_dr/line_for_taper.json'
    configs = [
        {'radiation_method': None, 'p0_correction': True, 'cavity_preserve_angle': False, 'beta_rtol': 2e-2, 'q_atol': 5e-4},
        {'radiation_method': 'kick_as_co', 'p0_correction': True, 'cavity_preserve_angle': False, 'beta_rtol': 2e-2, 'q_atol': 5e-4},
        {'radiation_method': 'kick_as_co', 'p0_correction': True, 'cavity_preserve_angle': True, 'beta_rtol': 1e-3, 'q_atol': 5e-4},
        {'radiation_method': 'scale_as_co', 'p0_correction': True, 'cavity_preserve_angle': True, 'beta_rtol': 2e-4, 'q_atol': 5e-4},
        {'radiation_method': 'full', 'p0_correction': False, 'cavity_preserve_angle': False, 'beta_rtol': 2e-2, 'q_atol': 5e-4},
        {'radiation_method': 'full', 'p0_correction': True, 'cavity_preserve_angle': False, 'beta_rtol': 2e-2, 'q_atol': 5e-4},
        {'radiation_method': 'full', 'p0_correction': True, 'cavity_preserve_angle': True, 'beta_rtol': 2e-4, 'q_atol': 5e-4},
    ]

    with open(filename, 'r') as f:
        line = xt.Line.from_dict(json.load(f))

    line.build_tracker()

    # Initial twiss (no radiation)
    line.configure_radiation(model=None)
    tw_no_rad = line.twiss(method='4d', freeze_longitudinal=True)

    assert tw_no_rad.radiation_method == None

    # Enable radiation
    line.configure_radiation(model='mean')
    # - Set cavity lags to compensate energy loss
    # - Taper magnet strengths
    line.compensate_radiation_energy_loss()

    for conf in configs:
        print(f'Running test with conf: {conf}')

        line.config.XTRACK_CAVITY_PRESERVE_ANGLE = conf['cavity_preserve_angle']

        extra_kwargs = {}
        if conf['radiation_method'] == 'kick_as_co' and conf['cavity_preserve_angle']:
            extra_kwargs['matrix_stability_tol'] = 0.1

        if conf['radiation_method'] == 'scale_as_co':
            extra_kwargs['matrix_stability_tol'] = 0.1
            extra_kwargs['use_full_inverse'] = True

        print('Twiss with radiation')
        # Twiss(es) with radiation
        tw = line.twiss(radiation_method=conf['radiation_method'],
                        eneloss_and_damping=True, **extra_kwargs)
        print('Done')

        assert tw.radiation_method == (conf['radiation_method'] or 'kick_as_co')

        # Check twiss at_s
        print('Twiss at_s')
        i_ele = len(tw.s)//3
        tws = line.twiss(at_s=tw.s[i_ele],
                        radiation_method=conf['radiation_method'],
                        eneloss_and_damping=True, **extra_kwargs)
        print('Done')

        line.config.XTRACK_CAVITY_PRESERVE_ANGLE = False

        if conf['p0_correction']:
            p0corr = 1 + tw.delta
        else:
            p0corr = 1

        # mask for taperable elemements
        tt = line.get_table().rows[:-1] # remove endpoint
        mask_taperable = (tt.element_type == 'Multipole') | (tt.element_type == 'DipoleEdge')
        assert np.sum(mask_taperable) == 17420

        delta_taper = line.attr['delta_taper']
        xo.assert_allclose(delta_taper[mask_taperable],
            0.5*(tw.delta[:-1] + tw.delta[1:])[mask_taperable], rtol=0, atol=1e-5)

        xo.assert_allclose(tw.delta[0], 0, rtol=0, atol=1e-5)
        xo.assert_allclose(tw.delta[-1], 0, rtol=0, atol=1e-5)

        xo.assert_allclose(tw.qx, tw_no_rad.qx, rtol=0, atol=conf['q_atol'])
        xo.assert_allclose(tw.qy, tw_no_rad.qy, rtol=0, atol=conf['q_atol'])

        xo.assert_allclose(tw.dqx, tw_no_rad.dqx, rtol=0, atol=1.5e-2*tw.qx)
        xo.assert_allclose(tw.dqy, tw_no_rad.dqy, rtol=0, atol=1.5e-2*tw.qy)

        xo.assert_allclose(tw.dqx, tw_no_rad.dqx, rtol=0, atol=1.5e-2*tw.qx)
        xo.assert_allclose(tw.y, tw_no_rad.y, rtol=0, atol=1e-6)

        xo.assert_allclose(tw.betx*p0corr, tw_no_rad.betx, rtol=conf['beta_rtol'], atol=0)
        xo.assert_allclose(tw.bety*p0corr, tw_no_rad.bety, rtol=conf['beta_rtol'], atol=0)

        xo.assert_allclose(tw.dx, tw.dx, rtol=0.0, atol=0.1e-3)

        xo.assert_allclose(tw.dy, tw.dy, rtol=0.0, atol=0.1e-3)

        xo.assert_allclose(tw.s[i_ele], tws.s, rtol=0, atol=1e-6)
        xo.assert_allclose(tw.x[i_ele], tws.x, rtol=0, atol=1e-6)
        xo.assert_allclose(tw.y[i_ele], tws.y, rtol=0, atol=1e-6)
        xo.assert_allclose(tw.betx[i_ele], tws.betx, rtol=1e-3, atol=0)
        xo.assert_allclose(tw.bety[i_ele], tws.bety, rtol=1e-3, atol=0)

        eneloss = tw.eneloss_turn
        assert eneloss/line.particle_ref.energy0 > 0.01
        xo.assert_allclose(
            line['rf'].voltage*np.sin((line['rf'].lag + line['rf'].lag_taper)/180*np.pi),
            eneloss/4, rtol=3e-5)
        xo.assert_allclose(
            line['rf1'].voltage*np.sin((line['rf'].lag + line['rf'].lag_taper)/180*np.pi),
            eneloss/4, rtol=3e-5)
        xo.assert_allclose(
            line['rf2a'].voltage*np.sin((line['rf'].lag + line['rf'].lag_taper)/180*np.pi),
            eneloss/4*0.6, rtol=3e-5)
        xo.assert_allclose(
            line['rf2b'].voltage*np.sin((line['rf'].lag + line['rf'].lag_taper)/180*np.pi),
            eneloss/4*0.4, rtol=3e-5)
        xo.assert_allclose(
            line['rf3'].voltage*np.sin((line['rf'].lag + line['rf'].lag_taper)/180*np.pi),
            eneloss/4, rtol=3e-5)



"""
Test implementation of the multipole shifts in solenoid slices
=============================================
Author(s): John P T Salvesen
Email:  john.salvesen@cern.ch
Date:   18-11-2024
"""
################################################################################
# Required Modules
################################################################################
import xtrack as xt
import numpy as np
import matplotlib.pyplot as plt

################################################################################
# User Parameters
################################################################################
N_SLICES    = int(1E2)
K0          = 1E-3
L_SOL       = 1
XING_RAD    = 1E-3

BETX        = 100E-3
BETY        = 1E-3

################################################################################
# Build Test Lines
################################################################################

########################################
# Build Environment
########################################
env     = xt.Environment(particle_ref = xt.Particles(p0c = 1E9))

########################################
# Line (beamline frame)
########################################
bl_components_in    = [env.new('bl_drift0', xt.Drift, length  = 1)]
bl_components_out   = [env.new('bl_drift1', xt.Drift, length  = 1)]

bl_components_sol = [
    env.new(f'bl_sol.{i}',  xt.Solenoid,
            length              = (L_SOL / N_SLICES),
            ks                  = 0,
            knl                 = [K0 * (L_SOL / N_SLICES), 0, 0],
            num_multipole_kicks = 1)
    for i in range(N_SLICES)]

bl_line    = env.new_line(
    components  = bl_components_in + bl_components_sol + bl_components_out)

########################################
# Line (horizontal rotated frame)
########################################
hrot_components_in = [
    env.new('hrot_drift0',       xt.Drift,       length  = 1),
    env.new('hshift_in',     xt.XYShift,     dx      = np.sin(XING_RAD) * L_SOL / 2),
    env.new('hrot_in',       xt.YRotation,   angle   = -np.rad2deg(XING_RAD))]

hrot_components_out = [
    env.new('hrot_out',      xt.YRotation,   angle   = np.rad2deg(XING_RAD)),
    env.new('hshift_out',    xt.XYShift,     dx      = np.sin(XING_RAD) * L_SOL / 2),
    env.new('hrot_drift1',       xt.Drift,       length  = 1)]

hrot_components_sol = [
    env.new(f'hrot_sol.{i}',  xt.Solenoid,
            length              = (L_SOL / N_SLICES) * np.cos(XING_RAD),
            ks                  = 0,
            knl                 = [K0 * (L_SOL / N_SLICES), 0, 0],
            num_multipole_kicks = 1,
            mult_rot_y_rad      = XING_RAD,
            mult_shift_x        = np.sin(XING_RAD) * L_SOL * (i/N_SLICES - 1/2))
    for i in range(N_SLICES)]

hrot_line    = env.new_line(
    components  = hrot_components_in + hrot_components_sol + hrot_components_out)

########################################
# Line (vertical rotated frame)
########################################
vrot_components_in = [
    env.new('vrot_drift0',   xt.Drift,       length  = 1),
    env.new('vshift_in',     xt.XYShift,     dy      = np.sin(XING_RAD) * L_SOL / 2),
    env.new('vrot_in',       xt.XRotation,   angle   = np.rad2deg(XING_RAD))]
# TODO: Minus sign difference here as still inconsistent definition with XRotation and YRotation
vrot_components_out = [
    env.new('vrot_out',      xt.XRotation,   angle   = -np.rad2deg(XING_RAD)),
    env.new('vshift_out',    xt.XYShift,     dy      = np.sin(XING_RAD) * L_SOL / 2),
    env.new('vrot_drift1',   xt.Drift,       length  = 1)]

vrot_components_sol = [
    env.new(f'vrot_sol.{i}',  xt.Solenoid,
            length              = (L_SOL / N_SLICES) * np.cos(XING_RAD),
            ks                  = 0,
            knl                 = [K0 * (L_SOL / N_SLICES), 0, 0],
            num_multipole_kicks = 1,
            mult_rot_x_rad      = XING_RAD,
            mult_shift_y        = np.sin(XING_RAD) * L_SOL * (i/N_SLICES - 1/2))
    for i in range(N_SLICES)]

vrot_line    = env.new_line(
    components  = vrot_components_in + vrot_components_sol + vrot_components_out)

################################################################################
# Comparisons
################################################################################
bl_twiss   = bl_line.twiss(
    method  = '4d',
    start   = xt.START,
    end     = xt.END,
    betx    = BETX,
    bety    = BETY)

hrot_twiss   = hrot_line.twiss(
    method  = '4d',
    start   = xt.START,
    end     = xt.END,
    betx    = BETX,
    bety    = BETY)

vrot_twiss   = vrot_line.twiss(
    method  = '4d',
    start   = xt.START,
    end     = xt.END,
    betx    = BETX,
    bety    = BETY)

bl_twiss.plot()
plt.title('Beamline Frame')
bl_twiss.plot('x y')
plt.title('Beamline Frame')

hrot_twiss.plot()
plt.title('Horizontal Rotated Frame')
hrot_twiss.plot('x y')
plt.title('Horizontal Rotated Frame')

vrot_twiss.plot()
plt.title('Vertical Rotated Frame')
vrot_twiss.plot('x y')
plt.title('Vertical Rotated Frame')

################################################################################
# Test Assertions
################################################################################
# Tolerances lower for derivative quantities (alfx, alfy, dpx, dpy)
assert np.isclose(bl_twiss['x'][-1],    hrot_twiss['x'][-1],    rtol = 1E-6)
assert np.isclose(bl_twiss['y'][-1],    hrot_twiss['y'][-1],    rtol = 1E-6)
assert np.isclose(bl_twiss['betx'][-1], hrot_twiss['betx'][-1], rtol = 1E-6)
assert np.isclose(bl_twiss['bety'][-1], hrot_twiss['bety'][-1], rtol = 1E-6)
assert np.isclose(bl_twiss['alfx'][-1], hrot_twiss['alfx'][-1], rtol = 1E-4)
assert np.isclose(bl_twiss['alfy'][-1], hrot_twiss['alfy'][-1], rtol = 1E-4)
assert np.isclose(bl_twiss['dx'][-1],   hrot_twiss['dx'][-1],   rtol = 1E-6)
assert np.isclose(bl_twiss['dy'][-1],   hrot_twiss['dy'][-1],   rtol = 1E-6)
assert np.isclose(bl_twiss['dpx'][-1],  hrot_twiss['dpx'][-1],  rtol = 1E-4)
assert np.isclose(bl_twiss['dpy'][-1],  hrot_twiss['dpy'][-1],  rtol = 1E-4)
assert np.isclose(bl_twiss['mux'][-1],  hrot_twiss['mux'][-1],  rtol = 1E-6)
assert np.isclose(bl_twiss['muy'][-1],  hrot_twiss['muy'][-1],  rtol = 1E-6)

assert np.isclose(bl_twiss['x'][-1],    vrot_twiss['x'][-1],    rtol = 1E-6)
assert np.isclose(bl_twiss['y'][-1],    vrot_twiss['y'][-1],    rtol = 1E-6)
assert np.isclose(bl_twiss['betx'][-1], vrot_twiss['betx'][-1], rtol = 1E-6)
assert np.isclose(bl_twiss['bety'][-1], vrot_twiss['bety'][-1], rtol = 1E-4)
assert np.isclose(bl_twiss['alfx'][-1], vrot_twiss['alfx'][-1], rtol = 1E-4)
assert np.isclose(bl_twiss['alfy'][-1], vrot_twiss['alfy'][-1], rtol = 1E-6)
assert np.isclose(bl_twiss['dx'][-1],   vrot_twiss['dx'][-1],   rtol = 1E-6)
assert np.isclose(bl_twiss['dy'][-1],   vrot_twiss['dy'][-1],   rtol = 1E-6)
assert np.isclose(bl_twiss['dpx'][-1],  vrot_twiss['dpx'][-1],  rtol = 1E-4)
assert np.isclose(bl_twiss['dpy'][-1],  vrot_twiss['dpy'][-1],  rtol = 1E-4)
assert np.isclose(bl_twiss['mux'][-1],  vrot_twiss['mux'][-1],  rtol = 1E-6)
assert np.isclose(bl_twiss['muy'][-1],  vrot_twiss['muy'][-1],  rtol = 1E-6)

########################################
# Show Plots
########################################
plt.show()



import json
import pathlib

import numpy as np

import xobjects as xo
import xtrack as xt
from xobjects.test_helpers import for_all_test_contexts

test_data_folder = pathlib.Path(
        __file__).parent.joinpath('../test_data').absolute()

@for_all_test_contexts
def test_match_orbit_bump(test_context):

    filename = test_data_folder.joinpath(
        'hllhc14_no_errors_with_coupling_knobs/line_b1.json')

    with open(filename, 'r') as fid:
        dct = json.load(fid)
    line = xt.Line.from_dict(dct)

    line.build_tracker(_context=test_context)

    tw_before = line.twiss()

    opt = line.match(
        start='mq.33l8.b1',
        end='mq.23l8.b1',
        init=tw_before.get_twiss_init(at_element='mq.33l8.b1'),
        vary=[
            xt.Vary(name='acbv30.l8b1', step=1e-10),
            xt.Vary(name='acbv28.l8b1', step=1e-10),
            xt.Vary(name='acbv26.l8b1', step=1e-10),
            xt.Vary(name='acbv24.l8b1', step=1e-10),
        ],
        targets=[
            # I want the vertical orbit to be at 3 mm at mq.28l8.b1 with zero angle
            xt.Target('y', at='mb.b28l8.b1', value=3e-3, tol=1e-4, scale=1),
            xt.Target('py', at='mb.b28l8.b1', value=0, tol=1e-6, scale=1000),
            # I want the bump to be closed
            xt.Target('y', at='mq.23l8.b1', value=tw_before['y', 'mq.23l8.b1'],
                      tol=1e-6, scale=1),
            xt.Target('py', at='mq.23l8.b1', value=tw_before['py', 'mq.23l8.b1'],
                      tol=1e-7, scale=1000),
        ]
    )
    assert len(opt.actions) == 1
    assert opt.actions[0].kwargs['_keep_initial_particles'] is True
    assert isinstance(opt.actions[0].kwargs['_initial_particles'], xt.Particles)

    tw = line.twiss()

    xo.assert_allclose(tw['y', 'mb.b28l8.b1'], 3e-3, atol=1e-4)
    xo.assert_allclose(tw['py', 'mb.b28l8.b1'], 0, atol=1e-6)
    xo.assert_allclose(tw['y', 'mq.23l8.b1'], tw_before['y', 'mq.23l8.b1'], atol=1e-6)
    xo.assert_allclose(tw['py', 'mq.23l8.b1'], tw_before['py', 'mq.23l8.b1'], atol=1e-7)
    xo.assert_allclose(tw['y', 'mq.33l8.b1'], tw_before['y', 'mq.33l8.b1'], atol=1e-6)
    xo.assert_allclose(tw['py', 'mq.33l8.b1'], tw_before['py', 'mq.33l8.b1'], atol=1e-7)

    # There is a bit of leakage in the horizontal plane (due to feed-down from sextupoles)
    xo.assert_allclose(tw['x', 'mb.b28l8.b1'], tw_before['x', 'mb.b28l8.b1'], atol=100e-6)
    xo.assert_allclose(tw['px', 'mb.b28l8.b1'], tw_before['px', 'mb.b28l8.b1'], atol=100e-7)
    xo.assert_allclose(tw['x', 'mq.23l8.b1'], tw_before['x', 'mq.23l8.b1'], atol=100e-6)
    xo.assert_allclose(tw['px', 'mq.23l8.b1'], tw_before['px', 'mq.23l8.b1'], atol=100e-7)
    xo.assert_allclose(tw['x', 'mq.33l8.b1'], tw_before['x', 'mq.33l8.b1'], atol=100e-6)
    xo.assert_allclose(tw['px', 'mq.33l8.b1'], tw_before['px', 'mq.33l8.b1'], atol=100e-7)

    # Now I match the bump including the horizontal plane
    # I start from scratch
    line.vars['acbv30.l8b1'] = 0
    line.vars['acbv28.l8b1'] = 0
    line.vars['acbv26.l8b1'] = 0
    line.vars['acbv24.l8b1'] = 0

    opt = line.match(
        start='mq.33l8.b1',
        end='mq.23l8.b1',
        init=tw_before.get_twiss_init(at_element='mq.33l8.b1'),
        vary=[
            xt.Vary(name='acbv30.l8b1', step=1e-10),
            xt.Vary(name='acbv28.l8b1', step=1e-10),
            xt.Vary(name='acbv26.l8b1', step=1e-10),
            xt.Vary(name='acbv24.l8b1', step=1e-10),
            xt.Vary(name='acbh27.l8b1', step=1e-10),
            xt.Vary(name='acbh25.l8b1', step=1e-10),
        ],
        targets=[
            # I want the vertical orbit to be at 3 mm at mq.28l8.b1 with zero angle
            xt.Target('y', at='mb.b28l8.b1', value=3e-3, tol=1e-4, scale=1),
            xt.Target('py', at='mb.b28l8.b1', value=0, tol=1e-6, scale=1000),
            # I want the bump to be closed
            xt.Target('y', at='mq.23l8.b1', value=tw_before['y', 'mq.23l8.b1'],
                    tol=1e-6, scale=1),
            xt.Target('py', at='mq.23l8.b1', value=tw_before['py', 'mq.23l8.b1'],
                    tol=1e-7, scale=1000),
            xt.Target('x', at='mq.23l8.b1', value=tw_before['x', 'mq.23l8.b1'],
                    tol=1e-6, scale=1),
            xt.Target('px', at='mq.23l8.b1', value=tw_before['px', 'mq.23l8.b1'],
                    tol=1e-7, scale=1000),
        ]
    )

    assert len(opt.actions) == 1
    assert opt.actions[0].kwargs['_keep_initial_particles'] is True
    assert isinstance(opt.actions[0].kwargs['_initial_particles'], xt.Particles)

    tw = line.twiss()
    xo.assert_allclose(tw['y', 'mb.b28l8.b1'], 3e-3, atol=1e-4)
    xo.assert_allclose(tw['py', 'mb.b28l8.b1'], 0, atol=1e-6)
    xo.assert_allclose(tw['y', 'mq.23l8.b1'], tw_before['y', 'mq.23l8.b1'], atol=1e-6)
    xo.assert_allclose(tw['py', 'mq.23l8.b1'], tw_before['py', 'mq.23l8.b1'], atol=1e-7)
    xo.assert_allclose(tw['y', 'mq.33l8.b1'], tw_before['y', 'mq.33l8.b1'], atol=1e-6)
    xo.assert_allclose(tw['py', 'mq.33l8.b1'], tw_before['py', 'mq.33l8.b1'], atol=1e-7)

    # There is a bit of leakage in the horizontal plane (due to feed-down from sextupoles)
    xo.assert_allclose(tw['x', 'mb.b28l8.b1'], tw_before['x', 'mb.b28l8.b1'], atol=50e-6)
    xo.assert_allclose(tw['px', 'mb.b28l8.b1'], tw_before['px', 'mb.b28l8.b1'], atol=2e-6)
    xo.assert_allclose(tw['x', 'mq.23l8.b1'], tw_before['x', 'mq.23l8.b1'], atol=1e-6)
    xo.assert_allclose(tw['px', 'mq.23l8.b1'], tw_before['px', 'mq.23l8.b1'], atol=1e-7)
    xo.assert_allclose(tw['x', 'mq.33l8.b1'], tw_before['x', 'mq.33l8.b1'], atol=1e-6)
    xo.assert_allclose(tw['px', 'mq.33l8.b1'], tw_before['px', 'mq.33l8.b1'], atol=1e-7)

    # Same match but with init provided through a kwargs
    # I start from scratch
    line.vars['acbv30.l8b1'] = 0
    line.vars['acbv28.l8b1'] = 0
    line.vars['acbv26.l8b1'] = 0
    line.vars['acbv24.l8b1'] = 0

    tini = tw_before.get_twiss_init(at_element='mq.33l8.b1')

    opt = line.match(
        start='mq.33l8.b1',
        end='mq.23l8.b1',
        betx=1, bety=1,
        x=tini.x, px=tini.px, y=tini.y, py=tini.py,
        zeta=tini.zeta, delta=tini.delta,
        vary=[
            xt.Vary(name='acbv30.l8b1', step=1e-10),
            xt.Vary(name='acbv28.l8b1', step=1e-10),
            xt.Vary(name='acbv26.l8b1', step=1e-10),
            xt.Vary(name='acbv24.l8b1', step=1e-10),
            xt.Vary(name='acbh27.l8b1', step=1e-10),
            xt.Vary(name='acbh25.l8b1', step=1e-10),
        ],
        targets=[
            # I want the vertical orbit to be at 3 mm at mq.28l8.b1 with zero angle
            xt.Target('y', at='mb.b28l8.b1', value=3e-3, tol=1e-4, scale=1),
            xt.Target('py', at='mb.b28l8.b1', value=0, tol=1e-6, scale=1000),
            # I want the bump to be closed
            xt.Target('y', at='mq.23l8.b1', value=tw_before['y', 'mq.23l8.b1'],
                    tol=1e-6, scale=1),
            xt.Target('py', at='mq.23l8.b1', value=tw_before['py', 'mq.23l8.b1'],
                    tol=1e-7, scale=1000),
            xt.Target('x', at='mq.23l8.b1', value=tw_before['x', 'mq.23l8.b1'],
                    tol=1e-6, scale=1),
            xt.Target('px', at='mq.23l8.b1', value=tw_before['px', 'mq.23l8.b1'],
                    tol=1e-7, scale=1000),
        ]
    )

    assert len(opt.actions) == 1
    assert opt.actions[0].kwargs['_keep_initial_particles'] is True
    assert isinstance(opt.actions[0].kwargs['_initial_particles'], xt.Particles)

    tw = line.twiss()
    xo.assert_allclose(tw['y', 'mb.b28l8.b1'], 3e-3, atol=1e-4)
    xo.assert_allclose(tw['py', 'mb.b28l8.b1'], 0, atol=1e-6)
    xo.assert_allclose(tw['y', 'mq.23l8.b1'], tw_before['y', 'mq.23l8.b1'], atol=1e-6)
    xo.assert_allclose(tw['py', 'mq.23l8.b1'], tw_before['py', 'mq.23l8.b1'], atol=1e-7)
    xo.assert_allclose(tw['y', 'mq.33l8.b1'], tw_before['y', 'mq.33l8.b1'], atol=1e-6)
    xo.assert_allclose(tw['py', 'mq.33l8.b1'], tw_before['py', 'mq.33l8.b1'], atol=1e-7)

    # There is a bit of leakage in the horizontal plane (due to feed-down from sextupoles)
    xo.assert_allclose(tw['x', 'mb.b28l8.b1'], tw_before['x', 'mb.b28l8.b1'], atol=50e-6)
    xo.assert_allclose(tw['px', 'mb.b28l8.b1'], tw_before['px', 'mb.b28l8.b1'], atol=2e-6)
    xo.assert_allclose(tw['x', 'mq.23l8.b1'], tw_before['x', 'mq.23l8.b1'], atol=1e-6)
    xo.assert_allclose(tw['px', 'mq.23l8.b1'], tw_before['px', 'mq.23l8.b1'], atol=1e-7)
    xo.assert_allclose(tw['x', 'mq.33l8.b1'], tw_before['x', 'mq.33l8.b1'], atol=1e-6)
    xo.assert_allclose(tw['px', 'mq.33l8.b1'], tw_before['px', 'mq.33l8.b1'], atol=1e-7)

def test_match_orbit_bump_with_weights():

    with open(test_data_folder /
              'hllhc14_no_errors_with_coupling_knobs/line_b1.json', 'r') as fid:
        dct = json.load(fid)
    line = xt.Line.from_dict(dct)

    line.build_tracker()

    for init_at in [xt.START, xt.END]:
        tw0 = line.twiss()
        opt = line.match(
            #verbose=True,
            solver='jacobian',
            # Portion of the beam line to be modified and initial conditions
            start='mq.33l8.b1',
            end='mq.17l8.b1',
            init_at=init_at,
            init=tw0,
            # Dipole corrector strengths to be varied
            vary=[
                xt.Vary(name='acbv32.l8b1', step=1e-10, weight=0.7),
                xt.Vary(name='acbv28.l8b1', step=1e-10, weight=0.3),
                xt.Vary(name='acbv26.l8b1', step=1e-10),
                xt.Vary(name='acbv24.l8b1', step=1e-10),
                xt.Vary(name='acbv22.l8b1', step=1e-10, limits=[-38e-6, 38e-6], weight=1000),
                xt.Vary(name='acbv18.l8b1', step=1e-10),
            ],
            targets=[
                # I want the vertical orbit to be at 3 mm at mq.28l8.b1 with zero angle
                xt.Target('y', at='mb.b26l8.b1', value=3e-3, tol=1e-4),
                xt.Target('py', at='mb.b26l8.b1', value=0, tol=1e-6),
                # I want the bump to be closed
                xt.Target('y', at='mq.17l8.b1', value=tw0, tol=1e-6),
                xt.Target('py', at='mq.17l8.b1', value=tw0, tol=1e-7, weight=1e3),
                xt.Target('y', at='mq.33l8.b1', value=tw0, tol=1e-6),
                xt.Target('py', at='mq.33l8.b1', value=tw0, tol=1e-7, weight=1e3),
                # I want to limit the negative excursion ot the bump
                xt.Target('y', xt.LessThan(-1e-3), at='mq.30l8.b1', tol=1e-6),
            ]
        )

        assert len(opt.actions) == 1
        assert opt.actions[0].kwargs['_keep_initial_particles'] is True
        assert isinstance(opt.actions[0].kwargs['_initial_particles'], xt.Particles)

        tw = line.twiss()

        xo.assert_allclose(tw['y', 'mq.33l8.b1'], 0, atol=1e-6, rtol=0)
        xo.assert_allclose(tw['y', 'mq.17l8.b1'], 0, atol=1e-6, rtol=0)
        xo.assert_allclose(tw['py', 'mq.17l8.b1'], 0, atol=1e-8, rtol=0)
        xo.assert_allclose(tw['py', 'mq.33l8.b1'], 0, atol=1e-6, rtol=0)

        xo.assert_allclose(tw['y', 'mb.b26l8.b1'], 3e-3, atol=1e-6, rtol=0)
        xo.assert_allclose(tw['py', 'mb.b26l8.b1'], 0, atol=1e-8, rtol=0)

        xo.assert_allclose(tw['y', 'mq.30l8.b1'], -1e-3, atol=1e-6, rtol=0)
        xo.assert_allclose(line.vars['acbv22.l8b1']._value, 38e-6, atol=0, rtol=0.02)

        # Extract last twiss done by optimizer
        last_data = opt._err._last_data
        action = list(last_data.keys())[0]
        last_twiss  = last_data[action]
        assert last_twiss.orientation == (
            'backward' if init_at == xt.END else 'forward')
        assert last_twiss.method == '6d'
        assert last_twiss.reference_frame == 'proper'

        targets = opt.targets
        assert targets[0].tar == ('y', 'mb.b26l8.b1')
        assert targets[0].weight == 10
        assert targets[1].tar == ('py', 'mb.b26l8.b1')
        assert targets[1].weight == 100
        assert targets[3].tar == ('py', 'mq.17l8.b1')
        assert targets[3].weight == 1000


@for_all_test_contexts
def test_match_orbit_bump_within_multiline(test_context):

    filename = test_data_folder.joinpath(
        'hllhc14_no_errors_with_coupling_knobs/line_b1.json')

    with open(filename, 'r') as fid:
        dct = json.load(fid)
    line = xt.Line.from_dict(dct)
    collider = xt.Environment(
        lines={'lhcb1': line}
    )

    collider.build_trackers(_context=test_context)

    tw_before = collider.twiss().lhcb1

    tw0 = collider.twiss()

    opt = collider.match(
        lines=['lhcb1'],
        start=['mq.33l8.b1'],
        end=['mq.23l8.b1'],
        init_at=xt.START,
        init=tw0,
        vary=[
            xt.Vary(name='acbv30.l8b1', step=1e-10),
            xt.Vary(name='acbv28.l8b1', step=1e-10),
            xt.Vary(name='acbv26.l8b1', step=1e-10),
            xt.Vary(name='acbv24.l8b1', step=1e-10),
        ],
        targets=[
            # I want the vertical orbit to be at 3 mm at mq.28l8.b1 with zero angle
            xt.Target('y', line='lhcb1', at='mb.b28l8.b1', value=3e-3, tol=1e-4, scale=1),
            xt.Target('py', line='lhcb1', at='mb.b28l8.b1', value=0, tol=1e-6, scale=1000),
            # I want the bump to be closed
            xt.Target('y', line='lhcb1', at='mq.23l8.b1', value=tw_before['y', 'mq.23l8.b1'],
                      tol=1e-6, scale=1),
            xt.Target('py', line='lhcb1', at='mq.23l8.b1', value=tw_before['py', 'mq.23l8.b1'],
                      tol=1e-7, scale=1000),
        ]
    )

    assert len(opt.actions) == 1
    assert len(opt.actions[0].kwargs['_keep_initial_particles']) == 1
    assert opt.actions[0].kwargs['_keep_initial_particles'][0] is True
    assert isinstance(opt.actions[0].kwargs['_initial_particles'][0], xt.Particles)

    tw_after_collider = collider.twiss()
    tw = tw_after_collider.lhcb1

    xo.assert_allclose(tw['y', 'mb.b28l8.b1'], 3e-3, atol=1e-4)
    xo.assert_allclose(tw['py', 'mb.b28l8.b1'], 0, atol=1e-6)
    xo.assert_allclose(tw['y', 'mq.23l8.b1'], tw_before['y', 'mq.23l8.b1'], atol=1e-6)
    xo.assert_allclose(tw['py', 'mq.23l8.b1'], tw_before['py', 'mq.23l8.b1'], atol=1e-7)
    xo.assert_allclose(tw['y', 'mq.33l8.b1'], tw_before['y', 'mq.33l8.b1'], atol=1e-6)
    xo.assert_allclose(tw['py', 'mq.33l8.b1'], tw_before['py', 'mq.33l8.b1'], atol=1e-7)

@for_all_test_contexts
def test_bump_step_and_smooth_inequalities(test_context):

    filename = test_data_folder.joinpath(
        'hllhc14_no_errors_with_coupling_knobs/line_b1.json')

    with open(filename, 'r') as fid:
        dct = json.load(fid)
    line = xt.Line.from_dict(dct)

    line.build_tracker(test_context)

    tw_before = line.twiss()

    GreaterThan = xt.GreaterThan
    LessThan = xt.LessThan

    tw0=line.twiss()

    opt = line.match(
        solve=False,
        solver='jacobian',
        # Portion of the beam line to be modified and initial conditions
        start='mq.33l8.b1',
        end='mq.17l8.b1',
        init_at=xt.START,
        init=tw0,
        # Dipole corrector strengths to be varied
        vary=[
            xt.Vary(name='acbv28.l8b1', step=1e-10),
            xt.Vary(name='acbv26.l8b1', step=1e-10),
            xt.Vary(name='acbv24.l8b1', step=1e-10),
            xt.Vary(name='acbv22.l8b1', step=1e-10),
        ],
        targets=[
            xt.Target('y', GreaterThan(2.7e-3), at='mb.b26l8.b1'),
            xt.Target('y', GreaterThan(2.7e-3), at='mb.b25l8.b1'),
            xt.Target('y', at='mq.24l8.b1', value=xt.LessThan(3e-3)),
            xt.Target('y', at='mq.26l8.b1', value=xt.LessThan(6e-3)),
            xt.TargetSet(['y', 'py'], at='mq.17l8.b1', value=tw0),
        ]
    )

    assert len(opt.actions) == 1
    assert opt.actions[0].kwargs['_keep_initial_particles'] is True
    assert isinstance(opt.actions[0].kwargs['_initial_particles'], xt.Particles)

    # Check freeze
    opt.step(1)
    ts = opt.target_status(ret=True)
    assert ts.tol_met[2] == False
    assert ts.tol_met[4] == False
    assert ts.tol_met[5] == False

    opt.targets[2].freeze()
    opt.targets[5].freeze()

    ts = opt.target_status(ret=True)
    assert ts.tol_met[2] == True
    assert ts.tol_met[4] == False
    assert ts.tol_met[5] == True

    opt.targets[2].unfreeze()
    opt.targets[5].unfreeze()

    ts = opt.target_status(ret=True)
    assert ts.tol_met[2] == False
    assert ts.tol_met[4] == False
    assert ts.tol_met[5] == False

    opt.solve()

    tw = line.twiss()
    assert tw['y', 'mb.b26l8.b1'] > 2.7e-3
    assert tw['y', 'mb.b25l8.b1'] > 2.7e-3
    assert tw['y', 'mq.24l8.b1'] < 3e-3 + 1e-6
    xo.assert_allclose(tw['y', 'mq.17l8.b1'], tw_before['y', 'mq.17l8.b1'], rtol=0, atol=1e-7)
    xo.assert_allclose(tw['py', 'mq.17l8.b1'], tw_before['py', 'mq.17l8.b1'], rtol=0, atol=1e-9)

    assert isinstance(opt.targets[0].value, xt.GreaterThan)
    assert isinstance(opt.targets[1].value, xt.GreaterThan)
    assert isinstance(opt.targets[2].value, xt.LessThan)

    assert opt.targets[0].value._value == 0
    assert opt.targets[1].value._value == 0
    assert opt.targets[2].value._value == 0

    assert opt.targets[0].value.mode == 'step'
    assert opt.targets[1].value.mode == 'step'
    assert opt.targets[2].value.mode == 'step'

    assert opt.targets[0].value.lower == 2.7e-3
    assert opt.targets[1].value.lower == 2.7e-3
    assert opt.targets[2].value.upper == 3e-3


    # Test mode smooth
    # Remove the bump

    for kk in ['acbv28.l8b1', 'acbv26.l8b1', 'acbv24.l8b1', 'acbv22.l8b1']:
        line.vars[kk] = 0

    tw_before = line.twiss()
    assert tw_before['y', 'mb.b26l8.b1'] < 1e-7
    assert tw_before['y', 'mb.b25l8.b1'] < 1e-7

    tw0 = line.twiss()
    opt = line.match(
        solve=False,
        solver='jacobian',
        # Portion of the beam line to be modified and initial conditions
        start='mq.33l8.b1',
        end='mq.17l8.b1',
        init_at=xt.START,
        init=tw0,
        # Dipole corrector strengths to be varied
        vary=[
            xt.Vary(name='acbv28.l8b1', step=1e-10),
            xt.Vary(name='acbv26.l8b1', step=1e-10),
            xt.Vary(name='acbv24.l8b1', step=1e-10),
            xt.Vary(name='acbv22.l8b1', step=1e-10),
        ],
        targets=[
            xt.Target('y', GreaterThan(2.7e-3, mode='smooth', sigma_rel=0.05), at='mb.b26l8.b1'),
            xt.Target('y', GreaterThan(2.7e-3, mode='smooth'), at='mb.b25l8.b1'),
            xt.Target('y', at='mq.24l8.b1', value=xt.LessThan(3e-3, mode='smooth', sigma_rel=0.04)),
            xt.Target('y', at='mq.26l8.b1', value=xt.LessThan(6e-3, mode='smooth')),
            xt.TargetSet(['y', 'py'], at='mq.17l8.b1', value=tw0),
        ]
    )

    assert len(opt.actions) == 1
    assert opt.actions[0].kwargs['_keep_initial_particles'] is True
    assert isinstance(opt.actions[0].kwargs['_initial_particles'], xt.Particles)

    # Check freeze
    opt.step(1)
    ts = opt.target_status(ret=True)
    assert ts.tol_met[0] == False
    assert ts.tol_met[1] == False
    assert ts.tol_met[5] == False

    opt.targets[0].freeze()
    opt.targets[5].freeze()

    ts = opt.target_status(ret=True)
    assert ts.tol_met[0] == True
    assert ts.tol_met[1] == False
    assert ts.tol_met[5] == True

    opt.targets[0].unfreeze()
    opt.targets[5].unfreeze()

    ts = opt.target_status(ret=True)
    assert ts.tol_met[0] == False
    assert ts.tol_met[1] == False
    assert ts.tol_met[5] == False

    opt.solve()

    tw = line.twiss()

    assert tw['y', 'mb.b26l8.b1'] > 2.7e-3 - 3e-6
    assert tw['y', 'mb.b25l8.b1'] > 2.7e-3 - 3e-6
    assert tw['y', 'mq.24l8.b1'] < 3e-3 + 3e-6
    assert tw['y', 'mq.26l8.b1'] < 6e-3 + 3e-6
    xo.assert_allclose(tw['y', 'mq.17l8.b1'], tw_before['y', 'mq.17l8.b1'], rtol=0, atol=1e-7)
    xo.assert_allclose(tw['py', 'mq.17l8.b1'], tw_before['py', 'mq.17l8.b1'], rtol=0, atol=1e-9)

    assert isinstance(opt.targets[0].value, xt.GreaterThan)
    assert isinstance(opt.targets[1].value, xt.GreaterThan)
    assert isinstance(opt.targets[2].value, xt.LessThan)
    assert isinstance(opt.targets[3].value, xt.LessThan)

    assert opt.targets[0].value._value == 0
    assert opt.targets[1].value._value == 0
    assert opt.targets[2].value._value == 0
    assert opt.targets[3].value._value == 0

    assert opt.targets[0].value.mode == 'smooth'
    assert opt.targets[1].value.mode == 'smooth'
    assert opt.targets[2].value.mode == 'smooth'
    assert opt.targets[3].value.mode == 'smooth'

    assert opt.targets[0].value.lower == 2.7e-3
    assert opt.targets[1].value.lower == 2.7e-3
    assert opt.targets[2].value.upper == 3e-3
    assert opt.targets[3].value.upper == 6e-3

    xo.assert_allclose(opt.targets[0].value.sigma, 0.05 * 2.7e-3, atol=0, rtol=1e-10)
    xo.assert_allclose(opt.targets[1].value.sigma, 0.01 * 2.7e-3, atol=0, rtol=1e-10)
    xo.assert_allclose(opt.targets[2].value.sigma, 0.04 * 3e-3, atol=0, rtol=1e-10)
    xo.assert_allclose(opt.targets[3].value.sigma, 0.01 * 6e-3, atol=0, rtol=1e-10)

    x_cut_norm = 1/16 + np.sqrt(33)/16
    poly = lambda x: 3 * x**3 - 2 * x**4

    # Check smooth target (GreaterThan)
    i_tar_gt = 0
    tar_gt = opt.targets[i_tar_gt]
    sigma_gt = tar_gt.value.sigma
    x0_gt = tar_gt.runeval()

    edge_test_gt = np.linspace(x0_gt - 3 * sigma_gt, x0_gt + 3 * sigma_gt, 100)

    residue_gt = edge_test_gt * 0
    for ii, xx in enumerate(edge_test_gt):
        tar_gt.value.lower = xx
        residue_gt[ii] =  opt._err()[i_tar_gt] / tar_gt.weight

    x_minus_edge_gt = x0_gt - edge_test_gt
    x_cut_gt = -x_cut_norm * sigma_gt

    mask_zero_gt = x_minus_edge_gt > 0
    assert np.all(residue_gt[mask_zero_gt] == 0)
    mask_linear_gt = x_minus_edge_gt < x_cut_gt
    xo.assert_allclose(residue_gt[mask_linear_gt],
        -x_minus_edge_gt[mask_linear_gt] - x_cut_norm * sigma_gt + sigma_gt*poly(x_cut_norm),
        atol=0, rtol=1e-10)
    mask_poly_gt = (~mask_zero_gt) & (~mask_linear_gt)
    xo.assert_allclose(residue_gt[mask_poly_gt],
        sigma_gt * poly(-x_minus_edge_gt[mask_poly_gt]/sigma_gt),
        atol=0, rtol=1e-10)

    # Check smooth target (LessThan)
    i_tar_lt = 2
    tar_lt = opt.targets[i_tar_lt]
    sigma_lt = tar_lt.value.sigma
    x0_lt = tar_lt.runeval()

    edge_test_lt = np.linspace(x0_lt - 3 * sigma_lt, x0_lt + 3 * sigma_lt, 100)

    residue_lt = edge_test_lt * 0
    for ii, xx in enumerate(edge_test_lt):
        tar_lt.value.upper = xx
        residue_lt[ii] =  opt._err()[i_tar_lt] / tar_lt.weight

    x_minus_edge_lt = x0_lt - edge_test_lt
    x_cut_lt = x_cut_norm * sigma_lt

    mask_zero_lt = x_minus_edge_lt < 0
    assert np.all(residue_lt[mask_zero_lt] == 0)
    mask_linear_lt = x_minus_edge_lt > x_cut_lt
    xo.assert_allclose(residue_lt[mask_linear_lt],
        x_minus_edge_lt[mask_linear_lt] - x_cut_norm * sigma_lt + sigma_lt*poly(x_cut_norm),
        atol=0, rtol=1e-10)
    mask_poly_lt = (~mask_zero_lt) & (~mask_linear_lt)
    xo.assert_allclose(residue_lt[mask_poly_lt],
        sigma_lt * poly(x_minus_edge_lt[mask_poly_lt]/sigma_lt),
        atol=0, rtol=1e-10)

@for_all_test_contexts
def test_match_bump_sets_implicit_end(test_context):

    line = xt.Line.from_json(test_data_folder /
                             'hllhc15_thick/lhc_thick_with_knobs.json')
    line.build_tracker(test_context)

    opt = line.match(
        start='mq.30l8.b1', end='mq.23l8.b1',
        betx=1, bety=1, y=0, py=0, # <-- conditions at start
        vary=xt.VaryList(['acbv30.l8b1', 'acbv28.l8b1', 'acbv26.l8b1', 'acbv24.l8b1'],
                        step=1e-10, limits=[-1e-3, 1e-3]),
        targets = [
            xt.TargetSet(y=3e-3, py=0, at='mb.b28l8.b1'),
            xt.TargetSet(y=0, py=0, at=xt.END)
        ])

    assert len(opt.actions) == 1
    assert opt.actions[0].kwargs['_keep_initial_particles'] is True
    assert isinstance(opt.actions[0].kwargs['_initial_particles'], xt.Particles)

    opt.tag(tag='matched')
    opt.reload(0)
    tw_before = line.twiss(method='4d')
    xo.assert_allclose(tw_before['y', 'mb.b28l8.b1'], 0, atol=1e-4)
    opt.reload(tag='matched')
    tw = line.twiss(method='4d')

    xo.assert_allclose(tw['y', 'mb.b28l8.b1'], 3e-3, atol=1e-4)
    xo.assert_allclose(tw['py', 'mb.b28l8.b1'], 0, atol=1e-6)
    xo.assert_allclose(tw['y', 'mq.23l8.b1'], tw_before['y', 'mq.23l8.b1'], atol=1e-6)
    xo.assert_allclose(tw['py', 'mq.23l8.b1'], tw_before['py', 'mq.23l8.b1'], atol=1e-7)
    xo.assert_allclose(tw['y', 'mq.33l8.b1'], tw_before['y', 'mq.33l8.b1'], atol=1e-6)
    xo.assert_allclose(tw['py', 'mq.33l8.b1'], tw_before['py', 'mq.33l8.b1'], atol=1e-7)

@for_all_test_contexts
def test_match_bump_sets_init_end(test_context):

    line = xt.Line.from_json(test_data_folder /
                             'hllhc15_thick/lhc_thick_with_knobs.json')
    line.build_tracker(test_context)

    opt = line.match(
        start='mq.30l8.b1', end='mq.23l8.b1',
        init_at=xt.END, betx=1, bety=1, y=0, py=0, # <-- conditions at end
        vary=xt.VaryList(['acbv30.l8b1', 'acbv28.l8b1', 'acbv26.l8b1', 'acbv24.l8b1'],
                        step=1e-10, limits=[-1e-3, 1e-3]),
        targets = [
            xt.TargetSet(y=3e-3, py=0, at='mb.b28l8.b1'),
            xt.TargetSet(y=0, py=0, at=xt.START)
    ])

    assert len(opt.actions) == 1
    assert opt.actions[0].kwargs['_keep_initial_particles'] is True
    assert isinstance(opt.actions[0].kwargs['_initial_particles'], xt.Particles)

    opt.tag(tag='matched')
    opt.reload(0)
    tw_before = line.twiss(method='4d')
    xo.assert_allclose(tw_before['y', 'mb.b28l8.b1'], 0, atol=1e-4)
    opt.reload(tag='matched')
    tw = line.twiss(method='4d')

    xo.assert_allclose(tw['y', 'mb.b28l8.b1'], 3e-3, atol=1e-4)
    xo.assert_allclose(tw['py', 'mb.b28l8.b1'], 0, atol=1e-6)
    xo.assert_allclose(tw['y', 'mq.23l8.b1'], tw_before['y', 'mq.23l8.b1'], atol=1e-6)
    xo.assert_allclose(tw['py', 'mq.23l8.b1'], tw_before['py', 'mq.23l8.b1'], atol=1e-7)
    xo.assert_allclose(tw['y', 'mq.33l8.b1'], tw_before['y', 'mq.33l8.b1'], atol=1e-6)
    xo.assert_allclose(tw['py', 'mq.33l8.b1'], tw_before['py', 'mq.33l8.b1'], atol=1e-7)

@for_all_test_contexts
def test_match_bump_sets_init_middle(test_context):

    line = xt.Line.from_json(test_data_folder /
                             'hllhc15_thick/lhc_thick_with_knobs.json')
    line.build_tracker(test_context)

    opt = line.match(
        start='mq.30l8.b1', end='mq.23l8.b1',
        init_at='mb.b28l8.b1', betx=1, bety=1, y=3e-3, py=0, # <-- conditions at point inside the range
        vary=xt.VaryList(['acbv30.l8b1', 'acbv28.l8b1', 'acbv26.l8b1', 'acbv24.l8b1'],
                        step=1e-10, limits=[-1e-3, 1e-3]),
        targets = [
            xt.TargetSet(y=0, py=0, at=xt.START),
            xt.TargetSet(y=0, py=0, at=xt.END)
    ])

    assert len(opt.actions) == 1
    assert opt.actions[0].kwargs['_keep_initial_particles'] is True
    assert opt.actions[0].kwargs['_initial_particles'] is None # due to init in the middle

    opt.tag(tag='matched')
    opt.reload(0)
    tw_before = line.twiss(method='4d')
    xo.assert_allclose(tw_before['y', 'mb.b28l8.b1'], 0, atol=1e-4)
    opt.reload(tag='matched')
    tw = line.twiss(method='4d')

    xo.assert_allclose(tw['y', 'mb.b28l8.b1'], 3e-3, atol=1e-4)
    xo.assert_allclose(tw['py', 'mb.b28l8.b1'], 0, atol=1e-6)
    xo.assert_allclose(tw['y', 'mq.23l8.b1'], tw_before['y', 'mq.23l8.b1'], atol=1e-6)
    xo.assert_allclose(tw['py', 'mq.23l8.b1'], tw_before['py', 'mq.23l8.b1'], atol=1e-7)
    xo.assert_allclose(tw['y', 'mq.33l8.b1'], tw_before['y', 'mq.33l8.b1'], atol=1e-6)
    xo.assert_allclose(tw['py', 'mq.33l8.b1'], tw_before['py', 'mq.33l8.b1'], atol=1e-7)

@for_all_test_contexts
def test_match_bump_sets_init_table(test_context):

    line = xt.Line.from_json(test_data_folder /
                             'hllhc15_thick/lhc_thick_with_knobs.json')
    line.build_tracker(test_context)

    tw0 = line.twiss(method='4d')
    opt = line.match(
        start='mq.30l8.b1', end='mq.23l8.b1',
        init=tw0, init_at=xt.END, # <-- Boundary conditions from table
        vary=xt.VaryList(['acbv30.l8b1', 'acbv28.l8b1', 'acbv26.l8b1', 'acbv24.l8b1'],
                        step=1e-10, limits=[-1e-3, 1e-3]),
        targets = [
            xt.TargetSet(y=3e-3, py=0, at='mb.b28l8.b1'),
            xt.TargetSet(['y', 'py'], value=tw0, at=xt.START) # <-- Target from table
        ])

    assert len(opt.actions) == 1
    assert opt.actions[0].kwargs['_keep_initial_particles'] is True
    assert isinstance(opt.actions[0].kwargs['_initial_particles'], xt.Particles)

    opt.tag(tag='matched')
    opt.reload(0)
    tw_before = line.twiss(method='4d')
    xo.assert_allclose(tw_before['y', 'mb.b28l8.b1'], 0, atol=1e-4)
    opt.reload(tag='matched')
    tw = line.twiss(method='4d')

    xo.assert_allclose(tw['y', 'mb.b28l8.b1'], 3e-3, atol=1e-4)
    xo.assert_allclose(tw['py', 'mb.b28l8.b1'], 0, atol=1e-6)
    xo.assert_allclose(tw['y', 'mq.23l8.b1'], tw_before['y', 'mq.23l8.b1'], atol=1e-6)
    xo.assert_allclose(tw['py', 'mq.23l8.b1'], tw_before['py', 'mq.23l8.b1'], atol=1e-7)
    xo.assert_allclose(tw['y', 'mq.33l8.b1'], tw_before['y', 'mq.33l8.b1'], atol=1e-6)
    xo.assert_allclose(tw['py', 'mq.33l8.b1'], tw_before['py', 'mq.33l8.b1'], atol=1e-7)

@for_all_test_contexts
def test_match_bump_common_elements(test_context):
    # Load a line and build a tracker
    collider = xt.Environment.from_json(test_data_folder /
                        'hllhc15_thick/hllhc15_collider_thick.json')
    collider.build_trackers(test_context)

    tw0 = collider.twiss(method='4d')

    opt = collider.match(
        lines=['lhcb1', 'lhcb2'],
        start=['e.ds.l5.b1', 'e.ds.l5.b2'],
        end=['s.ds.r5.b1', 's.ds.r5.b2'],
        init=tw0,
        vary=xt.VaryList([
            'acbxv1.r5', 'acbxv1.l5', # <-- common elements
            'acbyvs4.l5b1', 'acbrdv4.r5b1', 'acbcv5.l5b1', # <-- b1
            'acbyvs4.l5b2', 'acbrdv4.r5b2', 'acbcv5.r5b2', # <-- b2
            ],
            step=1e-10, limits=[-1e-3, 1e-3]),
        targets = [
            xt.TargetSet(y=0, py=10e-6, at='ip5', line='lhcb1'),
            xt.TargetSet(y=0, py=-10e-6, at='ip5', line='lhcb2'),
            xt.TargetSet(y=0, py=0, at=xt.END, line='lhcb1'),
            xt.TargetSet(y=0, py=0, at=xt.END, line='lhcb2')
        ])

    assert len(opt.actions) == 1
    assert len(opt.actions[0].kwargs['_keep_initial_particles']) == 2
    assert opt.actions[0].kwargs['_keep_initial_particles'][0] is True
    assert opt.actions[0].kwargs['_keep_initial_particles'][1] is True
    assert isinstance(opt.actions[0].kwargs['_initial_particles'][0], xt.Particles)
    assert isinstance(opt.actions[0].kwargs['_initial_particles'][1], xt.Particles)

    tw = collider.twiss()
    xo.assert_allclose(tw.lhcb1['y', 'ip5'], 0, rtol=0, atol=1e-9)
    xo.assert_allclose(tw.lhcb1['py', 'ip5'], 10e-6, rtol=0, atol=1e-9)
    xo.assert_allclose(tw.lhcb1['y', 's.ds.r5.b1'], 0, rtol=0, atol=1e-9)
    xo.assert_allclose(tw.lhcb1['py', 's.ds.r5.b1'], 0, rtol=0, atol=1e-9)
    xo.assert_allclose(tw.lhcb2['y', 'ip5'], 0, rtol=0, atol=1e-9)
    xo.assert_allclose(tw.lhcb2['py', 'ip5'], -10e-6, rtol=0, atol=1e-10)
    xo.assert_allclose(tw.lhcb2['y', 's.ds.r5.b2'], 0, rtol=0, atol=1e-9)
    xo.assert_allclose(tw.lhcb2['py', 's.ds.r5.b2'], 0, rtol=0, atol=1e-9)

@for_all_test_contexts
def test_match_bump_common_elements_callables_and_inequalities(test_context):
    # Load a line and build a tracker
    collider = xt.Environment.from_json(test_data_folder /
                        'hllhc15_thick/hllhc15_collider_thick.json')
    collider.build_trackers(test_context)

    tw0 = collider.twiss(method='4d')

    opt = collider.match(
        lines=['lhcb1', 'lhcb2'],
        start=['e.ds.l5.b1', 'e.ds.l5.b2'],
        end=['s.ds.r5.b1', 's.ds.r5.b2'],
        init=tw0,
        vary=xt.VaryList([
            'acbxv1.r5', 'acbxv1.l5', # <-- common elements
            'acbyvs4.l5b1', 'acbrdv4.r5b1', 'acbcv5.l5b1', # <-- b1
            'acbyvs4.l5b2', 'acbrdv4.r5b2', 'acbcv5.r5b2', # <-- b2
            ],
            step=1e-10, limits=[-1e-3, 1e-3]),
        targets = [
            xt.Target(y=0, at='ip5', line='lhcb1'),
            xt.Target('py', xt.GreaterThan(9e-6), at='ip5', line='lhcb1'), # <-- inequality
            xt.Target('py', xt.LessThan(  11e-6), at='ip5', line='lhcb1'), # <-- inequality
            xt.Target(y=0, at='ip5', line='lhcb2'),
            xt.Target(
                lambda tw: tw.lhcb1['py', 'ip5'] + tw.lhcb2['py', 'ip5'], value=0), # <-- callable
            xt.TargetSet(y=0, py=0, at=xt.END, line='lhcb1'),
            xt.TargetSet(y=0, py=0, at=xt.END, line='lhcb2')
        ])

    assert len(opt.actions) == 1
    assert len(opt.actions[0].kwargs['_keep_initial_particles']) == 2
    assert opt.actions[0].kwargs['_keep_initial_particles'][0] is True
    assert opt.actions[0].kwargs['_keep_initial_particles'][1] is True
    assert isinstance(opt.actions[0].kwargs['_initial_particles'][0], xt.Particles)
    assert isinstance(opt.actions[0].kwargs['_initial_particles'][1], xt.Particles)

    tw = collider.twiss()

    xo.assert_allclose(tw.lhcb1['y', 'ip5'], 0, rtol=0, atol=1e-9)
    xo.assert_allclose(tw.lhcb1['py', 'ip5'], 10e-6, rtol=0, atol=1.1e-6)
    xo.assert_allclose(tw.lhcb1['y', 's.ds.r5.b1'], 0, rtol=0, atol=1e-9)
    xo.assert_allclose(tw.lhcb1['py', 's.ds.r5.b1'], 0, rtol=0, atol=1e-9)
    xo.assert_allclose(tw.lhcb2['y', 'ip5'], 0, rtol=0, atol=1e-9)
    xo.assert_allclose(tw.lhcb2['py', 'ip5'] + tw.lhcb1['py', 'ip5'], 0, rtol=0, atol=1e-10)
    xo.assert_allclose(tw.lhcb2['y', 's.ds.r5.b2'], 0, rtol=0, atol=1e-9)
    xo.assert_allclose(tw.lhcb2['py', 's.ds.r5.b2'], 0, rtol=0, atol=1e-9)

@for_all_test_contexts
def test_match_bump_common_elements_targets_from_tables(test_context):
    # Load a line and build a tracker
    collider = xt.Environment.from_json(test_data_folder /
                        'hllhc15_thick/hllhc15_collider_thick.json')
    collider.build_trackers(test_context)

    tw0 = collider.twiss(method='4d')

    twb1 = collider.lhcb1.twiss(start='e.ds.l5.b1', end='s.ds.r5.b1', init=tw0.lhcb1)
    twb2 = collider.lhcb2.twiss(start='e.ds.l5.b2', end='s.ds.r5.b2', init=tw0.lhcb2)
    vars = collider.vars
    line_b1 = collider.lhcb1

    opt = collider.match(
        solve=False,
        vary=xt.VaryList([
            'acbxv1.r5', 'acbxv1.l5', # <-- common elements
            'acbyvs4.l5b1', 'acbrdv4.r5b1', 'acbcv5.l5b1', 'acbcv6.r5b1', # <-- b1
            'acbyvs4.l5b2', 'acbrdv4.r5b2', 'acbcv5.r5b2', 'acbcv6.l5b2'  # <-- b2
            ],
            step=1e-10, limits=[-1e-3, 1e-3]),
        targets = [
            # Targets from b1 twiss
            twb1.target(y=0, py=10e-6, at='ip5'),
            twb1.target(y=0, py=0, at=xt.END),
            # Targets from b2 twiss
            twb2.target(y=0, py=-10e-6, at='ip5'),
            twb2.target(['y', 'py'], at=xt.END), # <-- preserve
            # Targets from vars
            vars.target('acbxv1.l5', xt.LessThan(1e-3)),
            vars.target('acbxv1.l5', xt.GreaterThan(1e-6)),
            vars.target(lambda vv: vv['acbxv1.l5'] + vv['acbxv1.r5'], xt.LessThan(1e-9)),
            # Targets from line
            line_b1.target(lambda ll: ll['mcbrdv.4r5.b1']._xobject.ksl[0], xt.GreaterThan(1e-6)),
            line_b1.target(lambda ll: ll['mcbxfbv.a2r5/lhcb1']._xobject.ksl[0]
                                    + ll['mcbxfbv.a2l5/lhcb1']._xobject.ksl[0],
                                    xt.LessThan(1e-9)),
        ])
    opt.solve()

    assert len(opt.actions) == 7
    assert opt.actions[0].kwargs['_keep_initial_particles'] is True
    assert opt.actions[1].kwargs['_keep_initial_particles'] is True
    assert isinstance(opt.actions[0].kwargs['_initial_particles'], xt.Particles)
    assert isinstance(opt.actions[1].kwargs['_initial_particles'], xt.Particles)

    tw = collider.twiss()

    xo.assert_allclose(tw.lhcb1['y', 'ip5'], 0, rtol=0, atol=1e-9)
    xo.assert_allclose(tw.lhcb1['py', 'ip5'], 10e-6, rtol=0, atol=1.1e-6)
    xo.assert_allclose(tw.lhcb1['y', 's.ds.r5.b1'], 0, rtol=0, atol=1e-9)
    xo.assert_allclose(tw.lhcb1['py', 's.ds.r5.b1'], 0, rtol=0, atol=1e-9)
    xo.assert_allclose(tw.lhcb2['y', 'ip5'], 0, rtol=0, atol=1e-9)
    xo.assert_allclose(tw.lhcb2['py', 'ip5'] + tw.lhcb1['py', 'ip5'], 0, rtol=0, atol=1e-10)
    xo.assert_allclose(tw.lhcb2['y', 's.ds.r5.b2'], 0, rtol=0, atol=1e-9)
    xo.assert_allclose(tw.lhcb2['py', 's.ds.r5.b2'], 0, rtol=0, atol=1e-9)

@for_all_test_contexts
def test_match_bump_clone_and_ref_in_inequality(test_context):

    line = xt.Line.from_json(test_data_folder /
                            'hllhc14_no_errors_with_coupling_knobs/line_b1.json')
    line.build_tracker(test_context)

    GreaterThan = xt.GreaterThan
    LessThan = xt.LessThan

    tw0 = line.twiss()
    opt = line.match(
        name='bump',
        solve=False,
        solver='jacobian',
        # Portion of the beam line to be modified and initial conditions
        start='mq.33l8.b1',
        end='mq.17l8.b1',
        init=tw0, init_at=xt.START,
        # Dipole corrector strengths to be varied
        vary=[
            xt.Vary(name='acbv32.l8b1', step=1e-10, weight=0.7),
            xt.Vary(name='acbv28.l8b1', step=1e-10, weight=0.3),
            xt.Vary(name='acbv26.l8b1', step=1e-10),
            xt.Vary(name='acbv24.l8b1', step=1e-10),
            xt.Vary(name='acbv22.l8b1', step=1e-10, weight=1000),
            xt.Vary(name='acbv18.l8b1', step=1e-10),
        ],
        targets=[
            xt.Target('py', at='mb.b26l8.b1', value=0, tol=1e-6, weight=1e3),
            xt.Target('y', at='mb.b26l8.b1', value=3e-3, tol=1e-4),
            xt.Target('y', at='mq.17l8.b1', value=tw0, tol=1e-6),
            xt.Target('py', at='mq.17l8.b1', value=tw0, tol=1e-7, weight=1e3),
        ]
    )

    # Check target_mismatch
    assert opt.name == 'bump'
    ts = opt.target_status(ret=True)
    assert len(ts) == 4
    assert np.all(ts.tol_met == np.array([True, False, True, True]))
    tm = opt.target_mismatch(ret=True)
    assert len(tm) == 1
    assert tm.id[0] == 1

    opt.solve()

    # I want to limit the negative excursion ot the bump
    opt2 = opt.clone(name='limit',
        add_targets=[
            xt.Target('y', GreaterThan(-2e-3), at='mq.30l8.b1', tol=1e-6),
            xt.Target('y', GreaterThan(-1e-3), at='mq.30l8.b1', tol=1e-6)])
    opt2.solve()

    assert opt2.name == 'limit'
    assert len(opt2.targets) == 6
    tm = opt2.target_mismatch(ret=True)
    assert(len(tm) == 0)

    tw = line.twiss()

    assert np.isclose(tw['y', 'mq.33l8.b1'], 0, atol=1e-6, rtol=0)
    assert np.isclose(tw['y', 'mq.17l8.b1'], 0, atol=1e-6, rtol=0)
    assert np.isclose(tw['py', 'mq.17l8.b1'], 0, atol=1e-8, rtol=0)
    assert np.isclose(tw['py', 'mq.33l8.b1'], 0, atol=1e-6, rtol=0)

    assert np.isclose(tw['y', 'mb.b26l8.b1'], 3e-3, atol=1e-6, rtol=0)
    assert np.isclose(tw['py', 'mb.b26l8.b1'], 0, atol=1e-8, rtol=0)

    assert np.isclose(tw['y', 'mq.30l8.b1'], -1e-3, atol=1e-6, rtol=0)

    # Test variable in inequality
    line['myvar'] = -5e-3
    line['myvar2'] = 4e-3
    opt3 = opt2.clone(name='ineq',
        add_targets=[
            xt.Target('y', GreaterThan(line.ref['myvar']), at='mq.30l8.b1', tol=1e-6),
            xt.Target('y', LessThan(line.ref['myvar2']), at='mb.b26l8.b1', tol=1e-6)])

    assert len(opt3.target_mismatch(ret=True)) == 0
    assert opt3.target_status(ret=True).residue[-1] == 0

    line['myvar'] = -0.5e-3
    line['myvar2'] = 2e-3
    opt3.disable(target=1)
    assert len(opt3.target_mismatch(ret=True)) == 2
    xo.assert_allclose(opt3.target_mismatch(ret=True).residue[-2], -0.5e-3,
                    atol=1e-5, rtol=0)
    xo.assert_allclose(opt3.target_mismatch(ret=True).residue[-1], -1e-3,
                    atol=1e-5, rtol=0)

    opt3.solve()
    assert len(opt3.target_mismatch(ret=True)) == 1 # The disabled target
    assert opt3.target_mismatch(ret=True).id[0] == 1

    tw = line.twiss()

    assert np.isclose(tw['y', 'mq.33l8.b1'], 0, atol=1e-6, rtol=0)
    assert np.isclose(tw['y', 'mq.17l8.b1'], 0, atol=1e-6, rtol=0)
    assert np.isclose(tw['py', 'mq.17l8.b1'], 0, atol=1e-8, rtol=0)
    assert np.isclose(tw['py', 'mq.33l8.b1'], 0, atol=1e-6, rtol=0)

    assert np.isclose(tw['y', 'mb.b26l8.b1'], 2e-3, atol=1e-6, rtol=0)
    assert np.isclose(tw['py', 'mb.b26l8.b1'], 0, atol=1e-8, rtol=0)

    assert np.isclose(tw['y', 'mq.30l8.b1'], -0.5e-3, atol=1e-6, rtol=0)


def test_match_autogen_tags_and_clone_options():

    line = xt.Line.from_json(test_data_folder /
        "hllhc15_thick/lhc_thick_with_knobs.json")

    opt = line.match(
        start='mq.30l8.b1', end='mq.23l8.b1',
        betx=1, bety=1, y=0, py=0,
        vary=[xt.VaryList(['acbv30.l8b1', 'acbv28.l8b1',],
                        step=1e-10, limits=[-1e-3, 1e-3]),
            xt.VaryList(['acbv26.l8b1', 'acbv24.l8b1'],
                        step=1e-10, limits=[-1e-3, 1e-3], tag='mytag')],
        targets = [
            xt.TargetSet(y=3e-3, py=0, at='mb.b28l8.b1'),
            xt.TargetSet(y=0, py=0, at=xt.END)
        ])

    # Check autogenerated tags
    assert opt.targets[0].tag == 'mb.b28l8.b1_y'
    assert opt.targets[1].tag == 'mb.b28l8.b1_py'
    assert opt.targets[2].tag == 'END_y'
    assert opt.targets[3].tag == 'END_py'

    # Check target getitem
    assert opt.targets[0] == opt.targets['mb.b28l8.b1_y']
    assert opt.targets[1] == opt.targets['mb.b28l8.b1_py']
    assert opt.targets[2] == opt.targets['END_y']
    assert opt.targets[3] == opt.targets['END_py']

    assert opt.targets['END.*'][0] is opt.targets[2]
    assert opt.targets['END.*'][1] is opt.targets[3]

    assert np.all(
        opt.targets.status(ret=True).tag == opt.target_status(ret=True).tag)

    assert opt.vary['myt.*'][0] is opt.vary[2]
    assert opt.vary['myt.*'][1] is opt.vary[3]
    assert np.all(opt.vary.status(ret=True).tag == np.array(['', '', 'mytag', 'mytag']))
    assert np.all(opt.vary.status(ret=True).tag == opt.vary_status(ret=True).tag)
    assert np.all(opt.vary.status(ret=True).name == np.array(
        ['acbv30.l8b1', 'acbv28.l8b1', 'acbv26.l8b1', 'acbv24.l8b1']))

    opt1 = opt.clone(name='opt1')
    assert opt1.name == 'opt1'
    assert str(opt1.targets[0]) == str(opt.targets[0])
    assert str(opt1.targets[1]) == str(opt.targets[1])
    assert str(opt1.targets[2]) == str(opt.targets[2])
    assert str(opt1.targets[3]) == str(opt.targets[3])
    assert str(opt1.vary[0]) == str(opt.vary[0])
    assert str(opt1.vary[1]) == str(opt.vary[1])
    assert str(opt1.vary[2]) == str(opt.vary[2])
    assert str(opt1.vary[3]) == str(opt.vary[3])

    opt2 = opt.clone(name='opt2', remove_vary=True)
    assert len(opt2.vary) == 0
    assert str(opt2.targets[0]) == str(opt.targets[0])
    assert str(opt2.targets[1]) == str(opt.targets[1])
    assert str(opt2.targets[2]) == str(opt.targets[2])
    assert str(opt2.targets[3]) == str(opt.targets[3])

    opt3 = opt.clone(name='opt3', remove_targets=True,
                    add_targets=[xt.TargetSet(y=3e-3, py=0, at='mb.b28l8.b1')])
    assert len(opt3.targets) == 2
    assert opt.targets[0].tag == 'mb.b28l8.b1_y'
    assert opt.targets[1].tag == 'mb.b28l8.b1_py'
    assert str(opt3.vary[0]) == str(opt.vary[0])
    assert str(opt3.vary[1]) == str(opt.vary[1])
    assert str(opt3.vary[2]) == str(opt.vary[2])
    assert str(opt3.vary[3]) == str(opt.vary[3])

    opt4 = opt.clone(name='opt4', remove_vary=True, remove_targets=False,
                        add_vary=xt.VaryList(['acbv30.l8b1', 'acbv28.l8b1']),
                        )
    assert len(opt4.vary) == 2
    assert opt.vary[0].name == 'acbv30.l8b1'
    assert opt.vary[1].name == 'acbv28.l8b1'
